"""
This module contains utils for parallel batch calculation.
Main class is :py:class:`BatchManager` which is used to manage
parallel calculation

Main workflow is to add work with :py:meth:`BatchManager.add_work`
and consume results (:py:meth:`BatchManager.get_result`) until
:py:attr:`BatchManager.has_work` is evaluating to true

.. graphviz::

   digraph foo {
      "BatchManager" -> "BatchWorker"[arrowhead="crow"];
   }

"""

import logging
import multiprocessing
import os
import time
import traceback
import uuid
from contextlib import suppress
from enum import Enum
from queue import Empty, Queue
from threading import RLock, Timer
from typing import Any, Callable

__author__ = "Grzegorz Bokota"

from PartSegCore.plugins import register_if_need


class SubprocessOrder(Enum):
    """
    Commands for process to put in queue
    """

    kill = 1
    wait = 2
    cancel_job = 3


class BatchManager:
    """
    This class is used for manage pending works.
    It use :py:class:`.BatchWorker` for running calculation.

    :type task_queue: Queue
    :type order_queue: Queue
    :type result_queue: Queue
    :type calculation_dict: dict
    :type process_list: list[multiprocessing.Process]
    """

    def __init__(self):
        self.manager = multiprocessing.Manager()
        self.task_queue = self.manager.Queue()
        self.order_queue = self.manager.Queue()
        self.result_queue = self.manager.Queue()
        self.calculation_dict = self.manager.dict()
        self.number_off_available_process = 1
        self.number_off_process = 0
        self.number_off_alive_process = 0
        self.work_task = 0
        self.in_work = False
        self.process_list = []
        self.locker = RLock()

    def get_result(self) -> list[tuple[uuid.UUID, Any]]:
        """
        Clean result queue and return it as list

        :return: List of results as tuple where first element is uuid of job and second is
            function result or tuple with exception as first argument and second is traceback
        """
        res = []
        while not self.result_queue.empty():
            res.append(self.result_queue.get())
        self.work_task -= len(res)
        if self.work_task == 0:
            logging.debug("computation finished")
            Timer(0.1, self._change_process_num, args=[-self.number_off_available_process]).start()
        return res

    def add_work(self, individual_parameters_list: list, global_parameters, fun: Callable[[Any, Any], Any]) -> str:
        """
        This function add next works to internal structures.
        Number of works is length of ``individual_parameters_list``

        :param individual_parameters_list: list of individual parameters for fun.
            For each element ``fun`` will be called with element as first argument
        :param global_parameters: second argument of fun. If has field uuid then it is used as work uuid
        :param fun: two argument function which will be used to run calculation.
            First argument is task specific, second is const for whole work.
        :return: work uuid
        """
        self.calculation_dict[global_parameters.uuid] = global_parameters, fun
        self.work_task += len(individual_parameters_list)
        if hasattr(global_parameters, "uuid"):
            task_uuid = global_parameters.uuid
        else:
            task_uuid = uuid.uuid4()
        for el in individual_parameters_list:
            self.task_queue.put((el, task_uuid))
        if self.number_off_available_process > self.number_off_process:
            for _ in range(self.number_off_available_process - self.number_off_process):
                self._spawn_process()
        self.in_work = True
        return task_uuid

    def _spawn_process(self):
        with self.locker:
            process = multiprocessing.Process(
                target=spawn_worker, args=(self.task_queue, self.order_queue, self.result_queue, self.calculation_dict)
            )
            process.start()
            self.process_list.append(process)
            self.number_off_alive_process += 1
            self.number_off_process += 1

    @property
    def has_work(self) -> bool:
        """Check if Manager has pending or processed work and if all results are consumed"""
        return self.work_task > 0 or (not self.result_queue.empty())

    def kill_jobs(self):
        for p in self.process_list:
            p.terminate()

    def set_number_of_process(self, num: int):
        """
        Change number of workers which should be used for calculation

        :param num: target number of process
        """
        process_diff = num - self.number_off_available_process
        logging.debug("[set_number_of_process] process diff: %s", process_diff)
        self.number_off_available_process = num
        if not self.has_work:
            return
        self._change_process_num(process_diff)

    def _change_process_num(self, process_diff):
        if process_diff > 0:
            for _ in range(process_diff):
                self._spawn_process()
        else:
            for _ in range(-process_diff):
                logging.debug("[set_number_of_process] process kill")
                self.order_queue.put(SubprocessOrder.kill)
            with self.locker:
                self.number_off_process += process_diff
            self.join_all()

    def cancel_work(self, global_parameters):
        with suppress(KeyError):
            del self.calculation_dict[global_parameters.uuid]

    def join_all(self):
        logging.debug("Join begin %s %s", len(self.process_list), self.number_off_process)
        with self.locker:
            if len(self.process_list) > self.number_off_process:
                to_remove = []
                logging.debug("Process list start %s", self.process_list)
                for p in self.process_list:
                    if not p.is_alive():
                        p.join()
                        self.number_off_alive_process -= 1
                        to_remove.append(p)
                for p in to_remove:
                    self.process_list.remove(p)
                self.number_off_alive_process -= len(to_remove)
                logging.debug("Process list end %s", self.process_list)
            # FIXME self.number_off_alive_process,  self.number_off_process negative values
            if len(self.process_list) > self.number_off_process and len(self.process_list) > 0:
                logging.info(
                    "Wait on process, time %s, %s, %s, %s",
                    time.time(),
                    self.number_off_alive_process,
                    len(self.process_list),
                    self.number_off_process,
                )

                Timer(1, self.join_all).start()

    @property
    def finished(self):
        """Check if any process is running"""
        logging.debug(self.process_list)
        return len(self.process_list) == 0


class BatchWorker:
    """
    Worker spawned by :py:class:`BatchManager` instance

    :param task_queue: Queue with task data
    :param order_queue: Queue with additional orders (like kill)
    :param result_queue: Queue to put result
    :param calculation_dict: to store global parameters of task
    """

    def __init__(
        self,
        task_queue: Queue,
        order_queue: Queue,
        result_queue: Queue,
        calculation_dict: dict[uuid.UUID, tuple[Any, Callable[[Any, Any], Any]]],
    ):
        self.task_queue = task_queue
        self.order_queue = order_queue
        self.result_queue = result_queue
        self.calculation_dict = calculation_dict
        self.canceled_tasks = set()

    def calculate_task(self, val: tuple[Any, uuid.UUID]):
        """
        Calculate single task.
        ``val`` is tuple with two elements (task_data, uuid).
        function and global parameters are obtained from :py:attr:`.calculation_dict`
        """
        data, task_uuid = val
        calc = self.calculation_dict.get(task_uuid)
        if calc is None:
            self.result_queue.put((task_uuid, (-1, [SubprocessOrder.cancel_job])))
        global_data, fun = calc
        try:
            res = fun(data, global_data)
            self.result_queue.put((task_uuid, res))
        except Exception as e:  # pragma: no cover # pylint: disable=broad-except
            logging.exception("Exception in worker")
            self.result_queue.put((task_uuid, (-1, [(e, traceback.extract_tb(e.__traceback__))])))

    def run(self):
        """Worker main loop"""
        logging.debug("Process started %s", os.getpid())
        while True:
            if not self.order_queue.empty():
                with suppress(Empty):
                    order = self.order_queue.get_nowait()
                    logging.debug("Order message: %s", order)
                    if order == SubprocessOrder.kill:
                        break
            if not self.task_queue.empty():
                try:
                    task = self.task_queue.get_nowait()
                    self.calculate_task(task)
                except Empty:  # pragma: no cover
                    time.sleep(0.1)
                    continue
                except (MemoryError, OSError):  # pragma: no cover
                    pass
                except Exception as ex:  # pragma: no cover # pylint: disable=broad-except
                    logging.warning("Unsupported exception %s", ex)
            else:
                time.sleep(0.1)
        logging.info("Process %s ended", os.getpid())


def spawn_worker(task_queue: Queue, order_queue: Queue, result_queue: Queue, calculation_dict: dict[uuid.UUID, Any]):
    """
    Function for spawning worker. Designed as argument for :py:meth:`multiprocessing.Process`.

    :param task_queue: Queue with tasks
    :param order_queue: Queue with additional orders (like kill)
    :param result_queue: Queue for calculation result
    :param calculation_dict: dict with global parameters
    """
    try:
        register_if_need()
        with suppress(ImportError):
            from PartSeg.plugins import register_if_need as register

            register()
        worker = BatchWorker(task_queue, order_queue, result_queue, calculation_dict)
        worker.run()
    except Exception as e:  # pragma: no cover # pylint: disable=broad-except
        result_queue.put(("-1", (-1, [(e, traceback.extract_tb(e.__traceback__))])))



import numpy as np
import pytest

from PartSegCore.algorithm_describe_base import base_model_to_algorithm_property
from PartSegCore.segmentation import ROIExtractionAlgorithm
from PartSegCore.segmentation.algorithm_base import ROIExtractionResult, SegmentationLimitException
from PartSegCore.segmentation.restartable_segmentation_algorithms import (
    final_algorithm_list as restartable_list,
)
from PartSegCore.segmentation.restartable_segmentation_algorithms import (
    remove_object_touching_border,
)
from PartSegCore.segmentation.segmentation_algorithm import (
    CellFromNucleusFlow,
    ThresholdFlowAlgorithm,
)
from PartSegCore.segmentation.segmentation_algorithm import final_algorithm_list as algorithm_list
from PartSegCore.segmentation.utils import close_small_holes


def empty(*_):
    """Empty function to pass as a callback"""


@pytest.fixture(autouse=True)
def _fix_threshold_flow(monkeypatch):
    values = ThresholdFlowAlgorithm.get_default_values()
    values.threshold.values.core_threshold.values.threshold = 10
    values.threshold.values.base_threshold.values.threshold = 5

    def _param(self):
        return values

    monkeypatch.setattr(ThresholdFlowAlgorithm, "get_default_values", _param)

    values2 = CellFromNucleusFlow.get_default_values()
    values2.nucleus_threshold.values.threshold = 10
    values2.cell_threshold.values.threshold = 5

    def _param2(self):
        return values2

    monkeypatch.setattr(CellFromNucleusFlow, "get_default_values", _param2)


@pytest.mark.parametrize("algorithm", restartable_list + algorithm_list)
@pytest.mark.parametrize("masking", [True, False])
def test_segmentation_algorithm(image, algorithm: type[ROIExtractionAlgorithm], masking):
    assert algorithm.support_z() is True
    assert algorithm.support_time() is False
    assert isinstance(algorithm.get_steps_num(), int)
    instance = algorithm()
    instance.set_image(image)
    if masking:
        instance.set_mask(image.get_channel(0) > 0)
    if instance.__new_style__:
        # FIXME when migrate whole code
        instance.set_parameters(instance.get_default_values())
    else:
        instance.set_parameters(**instance.get_default_values())
    if not masking and "Need mask" in base_model_to_algorithm_property(instance.__argument_class__):
        with pytest.raises(SegmentationLimitException):
            instance.calculation_run(empty)
    else:
        res = instance.calculation_run(empty)
        assert isinstance(instance.get_info_text(), str)
        assert isinstance(res, ROIExtractionResult)
    instance.clean()


@pytest.mark.parametrize("ndim", [2, 3])
@pytest.mark.parametrize("dtype", [np.uint8, bool])
def test_close_small_holes(ndim, dtype):
    data = np.zeros((10,) * ndim, dtype=dtype)
    data[(slice(1, -1),) * ndim] = 1
    copy = data.copy()
    data[(slice(3, -3),) * ndim] = 0
    res = close_small_holes(data, 5**2)
    assert np.all(res == copy)


def test_remove_object_touching_border():
    data = np.zeros((10, 10), dtype=np.uint8)
    data[3:-3, 3:-3] = 1
    res = remove_object_touching_border(data)
    assert np.all(res == data)

    res = remove_object_touching_border(np.reshape(data, (1, 10, 10)))
    assert np.all(res == np.reshape(data, (1, 10, 10)))

    new_data = np.copy(data)
    new_data[:2] = 2

    res = remove_object_touching_border(new_data)
    assert np.all(res == data)

    res = remove_object_touching_border(np.reshape(new_data, (1, 10, 10)))
    assert np.all(res == np.reshape(data, (1, 10, 10)))



"""
THis module contains widgets used for error reporting. The report backed is sentry_.

.. _sentry: https://sentry.io
"""

import getpass
import io
import os
import pprint
import re
import traceback
import typing
from contextlib import suppress
from importlib.metadata import version

import numpy as np
import requests
import sentry_sdk
from napari.settings import get_settings
from napari.utils.theme import get_theme
from packaging.version import parse as parse_version
from qtpy.QtGui import QIcon
from qtpy.QtWidgets import (
    QApplication,
    QDialog,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QListWidget,
    QListWidgetItem,
    QMessageBox,
    QPushButton,
    QTextEdit,
    QTreeWidget,
    QTreeWidgetItem,
    QVBoxLayout,
    QWidget,
)
from sentry_sdk.utils import event_from_exception, exc_info_from_error
from traceback_with_variables import Format, print_exc

from PartSeg import __version__, state_store
from PartSeg.common_backend.python_syntax_highlight import Pylighter
from PartSegCore.io_utils import find_problematic_leafs
from PartSegCore.segmentation.algorithm_base import SegmentationLimitException
from PartSegCore.utils import numpy_repr

try:
    from qtpy import QT5
except ImportError:  # pragma: no cover
    QT5 = True

_EMAIL_REGEXP = re.compile(r"[\w+]+@\w+\.\w+")
_FEEDBACK_URL = "https://sentry.io/api/0/projects/{organization_slug}/{project_slug}/user-feedback/".format(
    organization_slug="cent", project_slug="partseg"
)
_napari_ge_5 = parse_version(version("napari")) >= parse_version("0.5.0a1")


def _print_traceback(exception, file_):
    while True:
        print_exc(exception, file_=file_, fmt=Format(custom_var_printers=[(np.ndarray, numpy_repr)]))
        if exception.__cause__ is not None:
            print("The above exception was the direct cause of the following exception:", file=file_)
            exception = exception.__cause__
        elif exception.__context__ is not None:
            print("During handling of the above exception, another exception occurred:", file=file_)
            exception = exception.__context__
        else:
            break


class ErrorDialog(QDialog):
    """
    Dialog to present user the exception information. User can send error report (possible to add custom information)
    """

    def __init__(self, exception: Exception, description: str, additional_notes: str = "", additional_info=None):
        super().__init__()
        self.exception = exception
        self.additional_notes = additional_notes
        self.send_report_btn = QPushButton("Report error")
        self.send_report_btn.setDisabled(not state_store.report_errors)
        self.create_issue_btn = QPushButton("Create issue")
        self.cancel_btn = QPushButton("Cancel")
        self.error_description = QTextEdit()
        if _napari_ge_5:
            theme = get_theme(get_settings().appearance.theme)
        else:
            theme = get_theme(get_settings().appearance.theme, as_dict=False)
        self._highlight = Pylighter(self.error_description.document(), "python", theme.syntax_style)
        self.traceback_summary = additional_info
        if additional_info is None:
            stream = io.StringIO()
            _print_traceback(exception, file_=stream)
            self.error_description.setText(stream.getvalue())
        elif isinstance(additional_info, traceback.StackSummary):
            self.error_description.setText("".join(additional_info.format()))
        elif isinstance(additional_info[1], traceback.StackSummary):
            self.error_description.setText("".join(additional_info[1].format()))
        self.error_description.append(str(exception))
        self.error_description.setReadOnly(True)
        self.additional_info = QTextEdit()
        self.contact_info = QLineEdit()
        self.user_name = QLineEdit()
        self.cancel_btn.clicked.connect(self.reject)
        self.send_report_btn.clicked.connect(self.send_report)
        self.create_issue_btn.clicked.connect(self.create_issue)

        self.desc = QLabel(description)
        self.desc.setWordWrap(True)

        self.setup_ui()

        if isinstance(additional_info, tuple):
            self.exception_tuple = additional_info[0], None
        else:
            exec_info = exc_info_from_error(exception)
            self.exception_tuple = event_from_exception(exec_info)

    def setup_ui(self):
        layout = QVBoxLayout()
        info_text = QLabel(
            "If you see these dialog it not means that you do something wrong. "
            "In such case you should see some message box not error report dialog."
        )
        info_text.setWordWrap(True)
        layout.addWidget(info_text)
        layout.addWidget(self.desc)
        layout.addWidget(self.error_description, 1)
        layout.addWidget(QLabel("Contact information"))
        layout.addWidget(self.contact_info)
        layout.addWidget(QLabel("User name"))
        layout.addWidget(self.user_name)
        layout.addWidget(QLabel("Additional information from user:"))
        layout.addWidget(self.additional_info)
        if not state_store.report_errors:
            layout.addWidget(
                QLabel(
                    "Sending reports was disabled by runtime flag. "
                    "You can report it manually by creating report on "
                    "https://github.com/4DNucleome/PartSeg/issues"
                )
            )
        btn_layout = QHBoxLayout()
        btn_layout.addWidget(self.cancel_btn)
        btn_layout.addWidget(self.create_issue_btn)
        btn_layout.addWidget(self.send_report_btn)
        layout.addLayout(btn_layout)
        self.setLayout(layout)

    if QT5:

        def exec(self):
            self.exec_()

    def exec_(self):
        """
        Check if dialog should be shown  base on :py:data:`state_store.show_error_dialog`.
        If yes then show dialog. Otherwise print exception traceback on stderr.
        """
        # TODO check if this check is needed
        if not state_store.show_error_dialog:
            print_exc(self.exception)
            return False
        super().exec_()
        return None

    def create_issue(self):
        """
        Create issue on github. This method is used when user disable error reporting.
        """
        import urllib.parse
        import webbrowser

        url = "https://github.com/4DNucleome/PartSeg/issues/new?"
        data = {
            "title": f"Error report from PartSeg `{self.exception!r}`",
            "body": f"This issue is created from PartSeg error dialog\n\n```"
            f"python\n{self.error_description.toPlainText()}\n```\n",
            "labels": "bug",
        }

        versions_dkt = {"PartSeg": __version__}

        with suppress(ModuleNotFoundError):
            from importlib.metadata import PackageNotFoundError, version

            for name in ["napari", "numpy", "SimpleITK", "PartSegData", "PartSegCore_compiled_backend"]:
                try:
                    versions_dkt[name] = version(name)
                except PackageNotFoundError:  # pragma: no cover   # noqa: PERF203
                    versions_dkt[name] = "not found"

        data["body"] += "Packages: \n```\n" + "\n".join(f"{k}=={v}" for k, v in versions_dkt.items()) + "\n```\n"

        webbrowser.open(f"{url}{urllib.parse.urlencode(data)}")

    def send_report(self):
        """
        Function with construct final error message and send it using sentry.
        """
        with sentry_sdk.new_scope() as scope:
            text = f"{self.desc.text()}\n\nVersion: {__version__}\n"
            if len(self.additional_notes) > 0:
                scope.set_extra("additional_notes", self.additional_notes)
            if len(self.additional_info.toPlainText()) > 0:
                scope.set_extra("user_information", self.additional_info.toPlainText())
            if len(self.contact_info.text()) > 0:
                scope.set_extra("contact", self.contact_info.text())
            event, hint = self.exception_tuple

            event["message"] = text
            if self.traceback_summary is not None:
                scope.set_extra("traceback", self.error_description.toPlainText())

            event_id = sentry_sdk.capture_event(event, hint=hint)
        if event_id is None:
            event_id = sentry_sdk.scope.Scope.last_event_id()

        if len(self.additional_info.toPlainText()) > 0:
            contact_text = self.contact_info.text()
            user_name = self.user_name.text()
            data = {
                "comments": self.additional_info.toPlainText(),
                "event_id": event_id,
                "email": contact_text if _EMAIL_REGEXP.match(contact_text) else "unknown@unknown.com",
                "name": user_name or get_user(),
            }

            with suppress(requests.exceptions.Timeout):
                r = requests.post(
                    url=_FEEDBACK_URL,
                    data=data,
                    headers={"Authorization": "DSN https://d4118280b73d4ee3a0222d0b17637687@sentry.io/1309302"},
                    timeout=3,
                )
                if r.status_code != 200:
                    data["email"] = "unknown@unknown.com"
                    data["name"] = get_user()
                    requests.post(
                        url=_FEEDBACK_URL,
                        data=data,
                        headers={"Authorization": "DSN https://d4118280b73d4ee3a0222d0b17637687@sentry.io/1309302"},
                        timeout=3,
                    )
        self.accept()


class ExceptionListItem(QListWidgetItem):
    """
    Element storing exception and showing basic information about it

    :param exception: exception or union of exception and traceback
    """

    # TODO Prevent from reporting disc error
    def __init__(
        self,
        exception: typing.Union[Exception, tuple[Exception, list]],
        parent: typing.Optional[QListWidget] = None,
    ):
        if isinstance(exception, Exception):
            traceback_summary = None
        else:
            exception, traceback_summary = exception
        if isinstance(exception, SegmentationLimitException):
            super().__init__(f"{exception}", parent, QListWidgetItem.ItemType.UserType)
        elif isinstance(exception, Exception):
            super().__init__(f"{type(exception)}: {exception}", parent, QListWidgetItem.ItemType.UserType)
            self.setToolTip("Double click for report")
        self.exception = exception
        self.additional_info = traceback_summary


class ExceptionList(QListWidget):
    """
    List to store exceptions
    """

    def __init__(self, parent=None):
        super().__init__(parent)
        self.itemDoubleClicked.connect(self.item_double_clicked)

    @staticmethod
    def item_double_clicked(el: QListWidgetItem):
        """
        if element clicked is :py:class:`ExceptionListItem` then open
        :py:class:`ErrorDialog` for reporting this error.

        This function is connected to :py:meth:`QListWidget.itemDoubleClicked`
        """
        if isinstance(el, ExceptionListItem) and not isinstance(el.exception, SegmentationLimitException):
            dial = ErrorDialog(el.exception, "Error during batch processing", additional_info=el.additional_info)
            dial.exec_()


class DataImportErrorDialog(QDialog):
    def __init__(
        self,
        errors: dict[str, typing.Union[Exception, list[tuple[str, dict]]]],
        parent: typing.Optional[QWidget] = None,
        text: str = "During import data part of the entries was filtered out",
    ):
        super().__init__(parent)
        self.setWindowTitle("Data import error")
        self.setWindowIcon(QIcon(":/icons/error.png"))
        self.setMinimumWidth(500)
        self.setMinimumHeight(500)
        self.setLayout(QVBoxLayout())
        self.setModal(True)
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().setSpacing(0)
        self.errors = errors
        self.error_view = QTreeWidget()
        self.layout().addWidget(QLabel(text))
        self.layout().addWidget(self.error_view)
        self.error_view.setHeaderLabels(["Keys", "Details"])
        for file_path, values in self.errors.items():
            file_item = QTreeWidgetItem(self.error_view, [file_path])
            file_item.setExpanded(True)
            file_item.setFirstColumnSpanned(True)
            if isinstance(values, Exception):
                QTreeWidgetItem(file_item, [str(values)]).setFirstColumnSpanned(True)
                continue
            for key, desc in values:
                problematic_entries = find_problematic_leafs(desc)
                item = QTreeWidgetItem(file_item, [key, str(problematic_entries[0]["__error__"]) + "..."])
                if len(problematic_entries) == 1:
                    QTreeWidgetItem(item, ["", pprint.pformat(problematic_entries[0])])
                if len(problematic_entries) > 1:
                    for entry in problematic_entries:
                        item2 = QTreeWidgetItem(item, ["", str(entry["__error__"])])
                        QTreeWidgetItem(item2, ["", pprint.pformat(entry)])

        close_btn = QPushButton("Close")
        close_btn.clicked.connect(self.accept)
        copy_btn = QPushButton("Copy to clipboard")
        copy_btn.clicked.connect(self._copy_to_clipboard)
        btn_layout = QHBoxLayout()
        btn_layout.addWidget(close_btn)
        btn_layout.addWidget(copy_btn)
        self.layout().addLayout(btn_layout)

    def _copy_to_clipboard(self):
        res = ""
        for file_path, values in self.errors.items():
            res += f"{file_path}\n"
            if isinstance(values, Exception):
                res += f"\t{values}\n"
                continue
            for key, desc in values:
                problematic_entries = find_problematic_leafs(desc)
                for entry in problematic_entries:
                    res += f"{key}: {entry['__error__']}\n{pprint.pformat(entry)}\n\n"

        QApplication.clipboard().setText(res)


class QMessageFromException(QMessageBox):
    """
    Specialized QMessageBox to provide information with attached exception

    """

    def __init__(self, icon, title, text, exception, standard_buttons=QMessageBox.StandardButton.Ok, parent=None):
        super().__init__(icon, title, text, standard_buttons, parent)
        self.exception = exception
        stream = io.StringIO()
        _print_traceback(exception, file_=stream)
        self.setDetailedText(stream.getvalue())

    @classmethod
    def critical(
        cls,
        parent=None,
        title="",
        text="",
        standard_buttons=QMessageBox.StandardButton.Ok,
        default_button=QMessageBox.StandardButton.NoButton,
        exception=None,
    ) -> QMessageBox.StandardButton:  # pylint: disable=arguments-differ
        ob = cls(
            icon=QMessageBox.Icon.Critical,
            title=title,
            text=text,
            exception=exception,
            parent=parent,
            standard_buttons=standard_buttons,
        )
        ob.setDefaultButton(default_button)
        return ob.exec_()

    @classmethod
    def information(
        cls,
        parent=None,
        title="",
        text="",
        standard_buttons=QMessageBox.StandardButton.Ok,
        default_button=QMessageBox.StandardButton.NoButton,
        exception=None,
    ) -> QMessageBox.StandardButton:  # pylint: disable=arguments-differ
        ob = cls(
            icon=QMessageBox.Icon.Information,
            title=title,
            text=text,
            exception=exception,
            parent=parent,
            standard_buttons=standard_buttons,
        )
        ob.setDefaultButton(default_button)
        return ob.exec_()

    @classmethod
    def question(
        cls,
        parent=None,
        title="",
        text="",
        standard_buttons=QMessageBox.StandardButton.Ok,
        default_button=QMessageBox.StandardButton.NoButton,
        exception=None,
    ) -> QMessageBox.StandardButton:  # pylint: disable=arguments-differ
        ob = cls(
            icon=QMessageBox.Icon.Question,
            title=title,
            text=text,
            exception=exception,
            parent=parent,
            standard_buttons=standard_buttons,
        )
        ob.setDefaultButton(default_button)
        return ob.exec_()

    @classmethod
    def warning(
        cls,
        parent=None,
        title="",
        text="",
        standard_buttons=QMessageBox.StandardButton.Ok,
        default_button=QMessageBox.StandardButton.NoButton,
        exception=None,
    ) -> QMessageBox.StandardButton:  # pylint: disable=arguments-differ
        ob = cls(
            icon=QMessageBox.Icon.Warning,
            title=title,
            text=text,
            exception=exception,
            parent=parent,
            standard_buttons=standard_buttons,
        )
        ob.setDefaultButton(default_button)
        return ob.exec_()


def get_user() -> str:
    try:
        return getpass.getuser()
    except ModuleNotFoundError:  # pragma: no cover
        # On windows `pwd` module is not available
        return os.getlogin()



import numpy as np

from PartSegCore.algorithm_describe_base import ROIExtractionProfile
from PartSegCore.project_info import AdditionalLayerDescription
from PartSegCore.segmentation.algorithm_base import ROIExtractionResult


class TestSegmentationResult:
    def test_update_alternative_names(self):
        res = ROIExtractionResult(
            roi=np.zeros((10, 10), dtype=np.uint8),
            parameters=ROIExtractionProfile(name="test", algorithm="test", values={}),
            additional_layers={
                "test1": AdditionalLayerDescription(np.zeros((10, 10), dtype=np.uint8), "image", "aa"),
                "test2": AdditionalLayerDescription(np.zeros((10, 10), dtype=np.uint8), "image", ""),
            },
        )
        assert res.additional_layers["test1"].name == "aa"
        assert res.additional_layers["test2"].name == "test2"



# pylint: disable=no-self-use

import os
from unittest.mock import MagicMock

import numpy as np
import pytest
from qtpy.QtCore import QMimeData, QPointF, Qt, QUrl
from qtpy.QtGui import QDropEvent
from qtpy.QtWidgets import QMessageBox

from PartSeg.common_gui.main_window import BaseMainMenu, BaseMainWindow, BaseSettings
from PartSeg.common_gui.waiting_dialog import ExecuteFunctionDialog
from PartSegCore.analysis import ProjectTuple
from PartSegCore.analysis.load_functions import LoadStackImage, load_dict
from PartSegCore.project_info import AdditionalLayerDescription
from PartSegImage import Image


class TestBaseMainWindow:
    def test_create(self, qtbot, tmp_path):
        main_window = BaseMainWindow(config_folder=tmp_path)
        qtbot.addWidget(main_window)

    def test_lack_of_config_folder(self, qtbot):
        with pytest.raises(ValueError, match="wrong config folder"):
            BaseMainWindow()

    def test_lack_of_config_with_settings(self, qtbot, tmp_path):
        main_window = BaseMainWindow(settings=BaseSettings(tmp_path))
        qtbot.addWidget(main_window)

    def test_toggle_console(self, qtbot, tmp_path):
        main_window = BaseMainWindow(config_folder=tmp_path)
        qtbot.addWidget(main_window)
        main_window.show()
        assert main_window.console is None
        main_window._toggle_console()
        assert main_window.console is not None
        assert main_window.console.isVisible()
        main_window._toggle_console()
        assert main_window.console is not None
        assert not main_window.console.isVisible()

    def test_toggle_multiple_files(self, qtbot, tmp_path):
        main_window = BaseMainWindow(config_folder=tmp_path)
        qtbot.addWidget(main_window)
        main_window.toggle_multiple_files()
        assert main_window.settings.get("multiple_files_widget")
        main_window.toggle_multiple_files()
        assert not main_window.settings.get("multiple_files_widget")

    def test_last_files(self, qtbot, monkeypatch, tmp_path, data_test_dir, part_settings):
        def mock_exec(self):
            self.thread_to_wait.run()
            return True

        monkeypatch.setattr(ExecuteFunctionDialog, "exec_", mock_exec)
        main_window = BaseMainWindow(settings=part_settings, load_dict=load_dict)
        main_window.main_menu = BaseMainMenu(main_window.settings, main_window)
        qtbot.addWidget(main_window)
        assert len(main_window.recent_file_menu.actions()) == 0
        main_window.settings.add_last_files(
            [os.path.join(data_test_dir, "stack1_components", "stack1_component1.tif")], LoadStackImage.get_name()
        )
        main_window.settings.add_last_files(
            [os.path.join(data_test_dir, "stack1_components", "stack1_component2.tif")], LoadStackImage.get_name()
        )
        main_window._refresh_recent()
        assert len(main_window.recent_file_menu.actions()) == 2
        assert (
            main_window.recent_file_menu.actions()[0].text()
            == f"{os.path.join(data_test_dir, 'stack1_components', 'stack1_component2.tif')}, "
            f"{LoadStackImage.get_name()}"
        )
        with qtbot.wait_signal(part_settings.image_changed):
            main_window.recent_file_menu.actions()[0].trigger()

    def test_get_colormaps(self, qtbot, part_settings):
        main_window = BaseMainWindow(settings=part_settings)
        qtbot.addWidget(main_window)
        assert len(main_window.get_colormaps()) == part_settings.image.channels

    @pytest.mark.windows_ci_skip
    def test_napari_viewer(self, qtbot, part_settings):
        main_window = BaseMainWindow(settings=part_settings)
        qtbot.addWidget(main_window)
        assert not main_window.viewer_list
        main_window.napari_viewer_show()
        qtbot.wait(50)
        assert len(main_window.viewer_list) == 1
        assert len(main_window.viewer_list[0].layers) == 2
        # with qtbot.wait_signal(main_window.viewer_list[0].window.qt_viewer.destroyed):
        main_window.viewer_list[0].close()
        qtbot.wait(50)
        assert not main_window.viewer_list

    def test_napari_viewer_additional_layers_empty(self, qtbot, part_settings, monkeypatch):
        main_window = BaseMainWindow(settings=part_settings)
        qtbot.addWidget(main_window)
        assert not main_window.viewer_list
        information_mock = MagicMock()
        monkeypatch.setattr(QMessageBox, "information", information_mock)
        main_window.additional_layers_show()
        assert not main_window.viewer_list
        information_mock.assert_called_once()

    @pytest.mark.windows_ci_skip
    @pytest.mark.pyside6_skip
    def test_napari_viewer_additional_layers(self, qtbot, part_settings, monkeypatch):
        main_window = BaseMainWindow(settings=part_settings)
        qtbot.addWidget(main_window)
        assert not main_window.viewer_list
        part_settings._additional_layers = {
            "layer1": AdditionalLayerDescription(np.zeros((10, 10)), "image"),
        }
        main_window.additional_layers_show()
        qtbot.wait(50)
        assert len(main_window.viewer_list) == 1
        assert len(main_window.viewer_list[0].layers) == 1
        # with qtbot.wait_signal(main_window.viewer_list[0].window.qt_viewer.destroyed):
        main_window.viewer_list[0].close()
        assert not main_window.viewer_list
        qtbot.wait(50)
        main_window.additional_layers_show(with_channels=True)
        qtbot.wait(50)
        assert len(main_window.viewer_list) == 1
        assert len(main_window.viewer_list[0].layers) == 3
        # with qtbot.wait_signal(main_window.viewer_list[0].window.qt_viewer.destroyed):
        main_window.viewer_list[0].close()
        qtbot.wait(50)
        assert not main_window.viewer_list

    def test_drop_files(self, qtbot, part_settings, data_test_dir, monkeypatch):
        def mock_exec(self):
            self.thread_to_wait.run()
            return True

        monkeypatch.setattr(ExecuteFunctionDialog, "exec_", mock_exec)

        main_window = BaseMainWindow(settings=part_settings, load_dict=load_dict)
        main_window.main_menu = BaseMainMenu(main_window.settings, main_window)
        qtbot.addWidget(main_window)
        data = QMimeData()
        data.setUrls([QUrl(f"file:/{data_test_dir}/stack1_components/stack1_component1.tif")])

        event = QDropEvent(QPointF(0, 0), Qt.MoveAction, data, Qt.NoButton, Qt.NoModifier)
        main_window.dropEvent(event)


class TestBaseMainMenu:
    def test_create(self, qtbot, tmp_path):
        main_menu = BaseMainMenu(BaseSettings(tmp_path), None)
        qtbot.addWidget(main_menu)

    def test_set_data_list_empty(self, qtbot, monkeypatch):
        warnings_mock = MagicMock()
        main_menu = BaseMainMenu(None, None)
        qtbot.addWidget(main_menu)
        monkeypatch.setattr(QMessageBox, "warning", warnings_mock)
        main_menu.set_data([])
        warnings_mock.assert_called_once()

    def test_none_data(self, qtbot, monkeypatch):
        warnings_mock = MagicMock()
        main_menu = BaseMainMenu(None, None)
        qtbot.addWidget(main_menu)
        monkeypatch.setattr(QMessageBox, "warning", warnings_mock)
        main_menu.set_data(None)
        warnings_mock.assert_called_once()

    def test_set_data_list(self, qtbot, part_settings, image, image2):
        main_window_mock = MagicMock()
        main_menu = BaseMainMenu(part_settings, main_window_mock)
        qtbot.addWidget(main_menu)
        data = [ProjectTuple("path", image), ProjectTuple("path2", image2)]
        main_menu.set_data(data)
        main_window_mock.multiple_files.add_states.assert_called_once()
        main_window_mock.multiple_files.add_states.assert_called_with(data)
        main_window_mock.multiple_files.setVisible.assert_called_with(True)

    def test_swap_time(self, qtbot, part_settings, image, monkeypatch):
        image = image.swap_time_and_stack()
        main_menu = BaseMainMenu(part_settings, None)
        qtbot.addWidget(main_menu)
        monkeypatch.setattr(QMessageBox, "question", lambda *_: QMessageBox.Yes)
        main_menu.set_data(ProjectTuple("path", image))
        assert part_settings.image.times == 1
        monkeypatch.setattr(QMessageBox, "question", lambda *_: QMessageBox.No)
        set_project_info_mock = MagicMock()
        monkeypatch.setattr(part_settings, "set_project_info", set_project_info_mock)
        main_menu.set_data(ProjectTuple("path", image))
        set_project_info_mock.assert_not_called()

    def test_time_and_stack_image(self, qtbot, part_settings, monkeypatch):
        image = Image(np.zeros((10, 10, 10, 10)), spacing=(1, 1, 1), axes_order="TZXY")
        main_menu = BaseMainMenu(part_settings, None)
        qtbot.addWidget(main_menu)
        warnings_mock = MagicMock()
        monkeypatch.setattr(QMessageBox, "question", lambda *_: QMessageBox.Yes)
        monkeypatch.setattr(QMessageBox, "warning", warnings_mock)
        main_menu.set_data(ProjectTuple("path", image))
        warnings_mock.assert_called_once()



from packaging.version import parse
from qtpy import QT_VERSION
from qtpy.QtCore import Qt, Signal
from qtpy.QtWidgets import QComboBox, QCompleter


class SearchComboBox(QComboBox):
    """
    ComboCox with completer for fast search in multiple options
    """

    if parse(QT_VERSION) < parse("5.14.0"):
        textActivated = Signal(str)  # pragma: no cover

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setEditable(True)
        self.completer_object = QCompleter()
        self.completer_object.setCaseSensitivity(Qt.CaseSensitivity.CaseInsensitive)
        self.completer_object.setCompletionMode(QCompleter.CompletionMode.PopupCompletion)
        self.completer_object.setFilterMode(Qt.MatchFlag.MatchContains)
        self.setCompleter(self.completer_object)
        self.setInsertPolicy(QComboBox.InsertPolicy.NoInsert)
        if parse(QT_VERSION) < parse("5.14.0"):  # pragma: no cover
            self.currentIndexChanged.connect(self._text_activated)

    def _text_activated(self):  # pragma: no cover
        self.textActivated.emit(self.currentText())

    def addItem(self, *args):
        super().addItem(*args)
        self.completer_object.setModel(self.model())

    def addItems(self, *args):
        super().addItems(*args)
        self.completer_object.setModel(self.model())



import itertools
from collections.abc import Iterator, MutableMapping
from typing import Any, ClassVar, Generic, TypeVar, Union

from qtpy.QtCore import QObject, Signal

from PartSeg.common_backend.abstract_class import QtMeta

T = TypeVar("T")
RemovableInfo = tuple[T, bool]


class PartiallyConstDict(QObject, MutableMapping, Generic[T], metaclass=QtMeta):
    """
    Base class for creating dict to mixin predefined and user defined variables.
    """

    item_added = Signal(object)
    """Signal with item added to dict"""
    item_removed = Signal(object)
    """Signal with item remove from dict"""
    const_item_dict: ClassVar[dict[str, Any]] = {}
    """Dict with non removable elements"""

    def __init__(self, editable_items):
        super().__init__()
        self.editable_items = editable_items
        self._order_dict = {
            name: i for i, name in enumerate(itertools.chain(self.const_item_dict.keys(), editable_items.keys()))
        }
        self._counter = len(self._order_dict)

    def __setitem__(self, key: str, value: Union[T, RemovableInfo]) -> None:
        if key in self.const_item_dict:
            raise ValueError("Cannot write base item")
        self.editable_items[key] = value[0] if isinstance(value, tuple) else value
        self._order_dict[key] = self._counter
        self._counter += 1
        self.item_added.emit(self.editable_items[key])

    def __len__(self) -> int:
        return len(self.editable_items) + len(self.const_item_dict)

    def __iter__(self) -> Iterator[str]:
        return itertools.chain(self.const_item_dict, self.editable_items)

    def __getitem__(self, key: str) -> RemovableInfo:
        try:
            if key in self.const_item_dict:
                return self.const_item_dict[key], False
            return self.editable_items[key], True
        except KeyError as e:
            raise KeyError(f"Element {key} not found") from e

    def __delitem__(self, key: str):
        if key in self.const_item_dict:
            raise ValueError(f"Cannot delete base item {key}")
        item = self.editable_items[key]
        del self.editable_items[key]
        del self._order_dict[key]
        self.item_removed.emit(item)

    def _refresh_order(self):
        """workaround for load data problem"""
        self._order_dict = {
            name: i for i, name in enumerate(itertools.chain(self.const_item_dict.keys(), self.editable_items.keys()))
        }
        self._counter = len(self._order_dict)

    def get_position(self, key: str) -> int:
        """
        Get item position as unique int. For sorting purpose

        :raise KeyError: if element not in dict
        """
        try:
            return self._order_dict[key]
        except KeyError:
            if key not in self:
                raise
            self._order_dict[key] = self._counter
            self._counter += 1
            return self._counter - 1



# pylint: disable=no-self-use
import numpy as np
import pytest

from PartSeg._roi_mask.main_window import ImageInformation, MainWindow
from PartSegCore.mask.io_functions import MaskProjectTuple
from PartSegCore.roi_info import ROIInfo


class TestMaskMainWindow:
    # @pytest.mark.skipif((platform.system() == "Linux") and CI_BUILD, reason="vispy problem")
    @pytest.mark.pyside_skip
    def test_opening(self, qtbot, tmpdir):
        main_window = MainWindow(tmpdir, initial_image=False)
        qtbot.addWidget(main_window)
        qtbot.wait(50)

    @pytest.mark.pyside_skip
    def test_change_theme(self, qtbot, tmpdir):
        main_window = MainWindow(tmpdir, initial_image=False)
        qtbot.addWidget(main_window)
        assert main_window.image_view.viewer.theme == "light"
        main_window.settings.theme_name = "dark"
        assert main_window.image_view.viewer.theme == "dark"

    @pytest.mark.pyside_skip
    def test_scale_bar(self, qtbot, tmpdir):
        main_window = MainWindow(tmpdir, initial_image=False)
        qtbot.addWidget(main_window)
        main_window._scale_bar_warning = False
        assert not main_window.image_view.viewer.scale_bar.visible
        main_window._toggle_scale_bar()
        assert main_window.image_view.viewer.scale_bar.visible

    def test_get_project_info(self, image, tmp_path):
        res = MainWindow.get_project_info(str(tmp_path / "test.tiff"), image)
        assert isinstance(res.roi_info, ROIInfo)
        assert res.roi_extraction_parameters == {}
        assert isinstance(res, MaskProjectTuple)

        roi = np.zeros(image.shape, dtype=np.uint8)
        roi[:, 2:-2] = 1
        res = MainWindow.get_project_info(str(tmp_path / "test.tiff"), image, ROIInfo(roi))
        assert res.roi_extraction_parameters == {1: None}

    @pytest.mark.pyside_skip
    def test_window_title(self, tmpdir, image, image2, qtbot):
        main_window = MainWindow(tmpdir, initial_image=False)
        qtbot.addWidget(main_window)
        assert main_window.windowTitle() == "PartSeg"
        main_window.settings.image = image
        assert main_window.windowTitle().replace("\\", "/") == "PartSeg: test_window_title0/test.tiff"

        image2.file_path = None
        main_window.settings.image = image2
        assert main_window.windowTitle() == "PartSeg"

        image.file_path = ""
        main_window.settings.image = image
        assert main_window.windowTitle() == "PartSeg: "


class TestImageInformation:
    def test_create(self, stack_settings, qtbot):
        info = ImageInformation(stack_settings)
        qtbot.addWidget(info)

    def test_multiple_files(self, stack_settings, qtbot):
        info = ImageInformation(stack_settings)
        qtbot.addWidget(info)

        assert info.multiple_files.isChecked()
        assert stack_settings.get("multiple_files_widget")

        stack_settings.set("multiple_files_widget", False)
        assert not info.multiple_files.isChecked()

        info.multiple_files.setChecked(True)
        assert stack_settings.get("multiple_files_widget")

    def test_sync_dirs(self, stack_settings, qtbot):
        info = ImageInformation(stack_settings)
        qtbot.addWidget(info)

        assert not info.sync_dirs.isChecked()
        assert not stack_settings.get("sync_dirs")

        stack_settings.set("sync_dirs", True)
        assert info.sync_dirs.isChecked()

        info.sync_dirs.setChecked(False)
        assert not stack_settings.get("sync_dirs")



from abc import ABC, abstractmethod
from collections.abc import MutableMapping
from copy import deepcopy
from dataclasses import dataclass, field
from textwrap import indent
from typing import Any, Callable, Optional

import numpy as np
from local_migrator import REGISTER, class_to_str

from PartSegCore.algorithm_describe_base import (
    AlgorithmDescribeBase,
    AlgorithmProperty,
    ROIExtractionProfile,
    base_model_to_algorithm_property,
)
from PartSegCore.image_operations import RadiusType
from PartSegCore.project_info import AdditionalLayerDescription
from PartSegCore.roi_info import ROIInfo
from PartSegCore.utils import BaseModel, numpy_repr
from PartSegImage import Channel, Image


def calculate_operation_radius(radius, spacing, gauss_type):
    if gauss_type == RadiusType.R2D and len(spacing) == 3:
        spacing = spacing[1:]
    base = min(spacing)
    if base != max(spacing):
        ratio = [x / base for x in spacing]
        return [radius / r for r in ratio]
    return radius


def dict_repr(dkt: MutableMapping) -> str:
    """
    calculate dict representation which use :py:func:`numpy_repr` for numpy representation.

    :param dict dkt: dict to be represented
    :return: string representation
    """
    res = []
    for k, v in dkt.items():
        if isinstance(v, MutableMapping):
            res.append(f"{k}: {dict_repr(v)}")
        elif isinstance(v, np.ndarray):
            res.append(f"{k}: {numpy_repr(v)}")
        else:
            res.append(f"{k}: {v!r}")
    return "{" + ", ".join(res) + "}"


@dataclass(frozen=True, repr=False)
class ROIExtractionResult:
    """
    Class to store results of roi extraction process.

    :ivar np.ndarray roi: Region of Interest represented as numpy array.
    :ivar ROIExtractionProfile parameters: parameters of roi extraction process.
    :ivar Dict[str,AdditionalLayerDescription] ~.additional_layers: additional layers returned by algorithm.
        Could be previewer using napari viewer.
    :ivar dict roi_annotation: Annotation for ROI. Currently displayed as tooltip
    :ivar Dict[str,np.ndarray] alternative_representation: Arrays with alternative representations of ROI.
    :ivar Optional[str] ~.file_path: information on which file roi extraction was performed.
    :ivar ROIInfo ~.roi_info: ROIInfo for current roi.
    :ivar Optional[np.ndarray] ~.points: array of points.
    """

    # TODO add alternative representation using dict mapping.
    roi: np.ndarray
    parameters: ROIExtractionProfile
    additional_layers: dict[str, AdditionalLayerDescription] = field(default_factory=dict)
    info_text: str = ""
    roi_annotation: dict = field(default_factory=dict)
    alternative_representation: dict[str, np.ndarray] = field(default_factory=dict)
    file_path: Optional[str] = None
    roi_info: Optional[ROIInfo] = None
    points: Optional[np.ndarray] = None

    def __post_init__(self):
        if "ROI" in self.alternative_representation:
            raise ValueError("alternative_representation field cannot contain field with ROI key")
        for key, value in self.additional_layers.items():
            if not value.name:
                value.name = key
        if self.roi_info is None:
            object.__setattr__(
                self,
                "roi_info",
                ROIInfo(roi=self.roi, annotations=self.roi_annotation, alternative=self.alternative_representation),
            )

    def __str__(self):  # pragma: no cover
        return (
            f"ROIExtractionResult(roi=[shape: {self.roi.shape}, dtype: {self.roi.dtype},"
            f" max: {np.max(self.roi)}], parameters={self.parameters},"
            f" additional_layers={list(self.additional_layers.keys())}, info_text={self.info_text},"
            f" alternative={dict_repr(self.alternative_representation)},"
            f" roi_annotation={dict_repr(self.roi_annotation)}"
        )

    def __repr__(self):  # pragma: no cover
        return (
            f"ROIExtractionResult(roi=[shape: {self.roi.shape}, dtype: {self.roi.dtype}, "
            f"max: {np.max(self.roi)}], parameters={self.parameters}, "
            f"additional_layers={list(self.additional_layers.keys())}, info_text={self.info_text},"
            f" alternative={dict_repr(self.alternative_representation)},"
            f" roi_annotation={dict_repr(self.roi_annotation)}"
        )


SegmentationResult = ROIExtractionResult


def report_empty_fun(_x, _y):  # pragma: no cover # skipcq: PTC-W0049
    pass


class AlgorithmInfo(BaseModel, arbitrary_types_allowed=True):
    algorithm_name: str
    parameters: Any
    image: Image


class ROIExtractionAlgorithm(AlgorithmDescribeBase, ABC):
    """
    Base class for all segmentation algorithm.

    :ivar Image ~.image: Image to process
    :ivar numpy.ndarray ~.channel: selected channel
    :ivar numpy.ndarray ~.segmentation: final segmentation
    :ivar numpy.ndarray ~.mask: mask limiting segmentation area
    """

    def __init__(self):
        super().__init__()
        self.image: Optional[Image] = None
        self.channel = None
        self.segmentation = None
        self._mask: Optional[np.ndarray] = None
        self.new_parameters: dict[str, Any] = {}

    def __repr__(self):  # pragma: no cover
        if self.mask is None:
            mask_info = "mask=None"
        elif isinstance(self.mask, np.ndarray):
            mask_info = (
                f"mask_dtype={self.mask.dtype}, mask_shape={self.mask.shape}, mask_unique={np.unique(self.mask)}"
            )
        else:
            mask_info = f"mask={self.mask}"
        return (
            f"{self.__class__.__module__}.{self.__class__.__name__}(\n"
            + indent(f"image={self.image!r},\n", " " * 4)
            + indent(f"channel={numpy_repr(self.channel)},\n{mask_info},", " " * 4)
            + indent(f"\nvalue={self.get_segmentation_profile().values!r})", " " * 4)
        )

    def clean(self):
        self.image = None
        self.segmentation = None
        self.channel = None
        self.mask = None

    @property
    def mask(self) -> Optional[np.ndarray]:
        if self._mask is not None and not self.support_time():
            return self.image.clip_array(self._mask, t=0)
        return self._mask

    @mask.setter
    def mask(self, val: Optional[np.ndarray]):
        if val is None:
            self._mask = None
            return
        self._mask = self.image.fit_mask_to_image(val)

    @classmethod
    @abstractmethod
    def support_time(cls):
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def support_z(cls):
        raise NotImplementedError

    def set_mask(self, mask):
        """Set mask which will limit segmentation area"""
        self.mask = mask

    def calculation_run_wrap(self, report_fun: Callable[[str, int], None]) -> ROIExtractionResult:
        try:
            return self.calculation_run(report_fun)
        except SegmentationLimitException:  # pragma: no cover
            raise
        except Exception as e:  # pragma: no cover
            parameters = self.get_segmentation_profile()
            image = self.image
            raise SegmentationException(
                AlgorithmInfo(algorithm_name=self.get_name(), parameters=parameters, image=image)
            ) from e

    @abstractmethod
    def calculation_run(self, report_fun: Callable[[str, int], None]) -> ROIExtractionResult:
        raise NotImplementedError

    @abstractmethod
    def get_info_text(self):
        raise NotImplementedError

    def get_channel(self, channel_idx):
        if self.support_time():
            return self.image.get_data_by_axis(c=channel_idx)
        if self.image.is_time:
            raise ValueError("This algorithm do not support time data")
        if isinstance(channel_idx, int) and self.image.channels <= channel_idx:
            raise SegmentationException(
                f"Image {self.image} has only {self.image.channels} when requested channel {channel_idx}"
            )
        if isinstance(channel_idx, str) and channel_idx not in self.image.channel_names:
            raise SegmentationException(
                f"Image {self.image} has only {self.image.channel_names} when requested channel '{channel_idx}'"
            )
        return self.image.get_data_by_axis(c=channel_idx, t=0)

    def set_image(self, image):
        self.image = image
        self.channel = None
        self.mask = None

    def set_parameters(self, _params=None, **kwargs):
        # FIXME when drop python 3.7 use positional only argument
        if _params is not None:
            if isinstance(_params, dict):
                kwargs = _params
            else:
                self.new_parameters = _params
                return
        if self.__new_style__:
            kwargs = REGISTER.migrate_data(class_to_str(self.__argument_class__), {}, kwargs)
            self.new_parameters = self.__argument_class__(**kwargs)  # pylint: disable=not-callable
            return

        base_names = [x.name for x in self.get_fields() if isinstance(x, AlgorithmProperty)]
        if set(base_names) != set(kwargs.keys()):
            missed_arguments = ", ".join(set(base_names).difference(set(kwargs.keys())))
            additional_arguments = ", ".join(set(kwargs.keys()).difference(set(base_names)))
            raise ValueError(f"Missed arguments {missed_arguments}; Additional arguments: {additional_arguments}")
        self.new_parameters = deepcopy(kwargs)

    def get_segmentation_profile(self) -> ROIExtractionProfile:
        return ROIExtractionProfile(name="", algorithm=self.get_name(), values=deepcopy(self.new_parameters))

    @staticmethod
    def get_steps_num():
        """Return number of algorithm steps if your algorithm report progress, else should return 0"""
        return 0

    @classmethod
    def get_channel_parameter_name(cls):
        if cls.__new_style__:
            fields = base_model_to_algorithm_property(cls.__argument_class__)
        else:
            fields = cls.get_fields()
        for el in fields:
            if el.value_type == Channel:
                return el.name
        raise ValueError("No channel defined")


class SegmentationLimitException(Exception):
    pass


class SegmentationException(Exception):
    pass


SegmentationAlgorithm = ROIExtractionAlgorithm  # rename backward compatibility



"""
This module contains base for the advanced window for PartSeg.
At this moment controlling colormaps tabs and developer PartSegCore
"""

import importlib
import logging
from contextlib import suppress

from qtpy.QtCore import QByteArray, Qt
from qtpy.QtGui import QCloseEvent
from qtpy.QtWidgets import (
    QCheckBox,
    QColorDialog,
    QComboBox,
    QDoubleSpinBox,
    QGridLayout,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QTabWidget,
    QVBoxLayout,
    QWidget,
)

from PartSeg import plugins, state_store
from PartSeg._roi_analysis.partseg_settings import PartSettings
from PartSeg.common_backend.base_settings import BaseSettings, ViewSettings
from PartSeg.common_gui.colormap_creator import PColormapCreator, PColormapList
from PartSeg.common_gui.dict_viewer import DictViewer
from PartSeg.common_gui.label_create import ColorShow, LabelChoose, LabelEditor
from PartSeg.common_gui.universal_gui_part import CustomDoubleSpinBox
from PartSegCore import plugins as core_plugins
from PartSegCore import register

RENDERING_LIST = ["iso_categorical", "translucent"]

RENDERING_MODE_NAME_STR = "rendering_mode"
SEARCH_ZOOM_FACTOR_STR = "search_zoom_factor"


class DevelopTab(QWidget):
    """
    Widget for developer utilities. Currently only contains button for reload algorithms and

    To enable it run program with a `--develop` flag.

    If you would like to use it for developing your own algorithm and modify some of the PartSeg class.
    Please protect this part of code with something like:

    >>> if tifffile.tifffile.TiffPage.__module__ != "PartSegImage.image_reader":

    This is taken from :py:mod:`PartSegImage.image_reader`
    """

    def __init__(self, parent=None):
        super().__init__(parent=parent)

        # noinspection PyArgumentList
        self.reload_btn = QPushButton("Reload algorithms", clicked=self.reload_algorithm_action)
        layout = QGridLayout(self)
        layout.addWidget(self.reload_btn, 0, 0)
        layout.setColumnStretch(1, 1)
        layout.setRowStretch(1, 1)
        self.setLayout(layout)

    def reload_algorithm_action(self):
        """Function for reload plugins and algorithms"""
        msg = "Reloading %s"
        for val in register.reload_module_list:
            logging.info(msg, val.__name__)
            importlib.reload(val)
        for el in plugins.get_plugins():
            logging.info(msg, el.__name__)
            importlib.reload(el)
        for el in core_plugins.get_plugins():
            logging.info(msg, el.__name__)
            importlib.reload(el)
        importlib.reload(register)
        importlib.reload(plugins)
        importlib.reload(core_plugins)
        plugins.register()
        core_plugins.register()
        if hasattr(self.parent().parent(), "reload_list"):
            for el in self.parent().parent().reload_list:
                el()


class MaskControl(QWidget):
    def __init__(self, settings: ViewSettings, parent=None):
        super().__init__(parent=parent)
        self.settings = settings
        self.color_picker = QColorDialog()
        self.color_picker.setWindowFlag(Qt.WindowType.Widget)
        self.color_picker.setOptions(
            QColorDialog.ColorDialogOption.DontUseNativeDialog | QColorDialog.ColorDialogOption.NoButtons
        )
        self.opacity_spin = QDoubleSpinBox(self)
        self.opacity_spin.setRange(0, 1)
        self.opacity_spin.setSingleStep(0.1)
        self.opacity_spin.setDecimals(2)
        self.change_mask_color_btn = QPushButton("Change mask color")
        self.current_mask_color_preview = ColorShow(
            self.settings.get_from_profile("mask_presentation_color", [255, 255, 255])
        )

        self.opacity_spin.setValue(self.settings.get_from_profile("mask_presentation_opacity", 1))

        self.current_mask_color_preview.setAutoFillBackground(True)
        self.change_mask_color_btn.clicked.connect(self.change_color)
        self.opacity_spin.valueChanged.connect(self.change_opacity)

        layout = QVBoxLayout()
        layout.addWidget(self.color_picker)
        layout2 = QHBoxLayout()
        layout2.addWidget(self.change_mask_color_btn)
        layout2.addWidget(self.current_mask_color_preview, 1)
        layout2.addWidget(QLabel("Mask opacity"))
        layout2.addWidget(self.opacity_spin)
        layout.addLayout(layout2)
        self.setLayout(layout)

    def set_color_preview(self, color):
        self.current_mask_color_preview.set_color(color)

    def change_color(self):
        color = self.color_picker.currentColor()
        color = (color.red(), color.green(), color.blue())
        self.settings.set_in_profile("mask_presentation_color", color)
        self.set_color_preview(color)

    def change_opacity(self):
        self.settings.set_in_profile("mask_presentation_opacity", self.opacity_spin.value())


class Appearance(QWidget):
    def __init__(self, settings: ViewSettings):
        super().__init__()
        self.settings = settings

        self.layout_list = QComboBox()
        self.layout_list.addItems(self.settings.theme_list())
        self.layout_list.setCurrentText(self.settings.theme_name)
        self.layout_list.currentIndexChanged.connect(self.change_theme)

        self.labels_render_cmb = QComboBox()
        self.labels_render_cmb.addItems(RENDERING_LIST)
        self._update_render_mode()
        self.labels_render_cmb.currentTextChanged.connect(self.change_render_mode)
        settings.connect_to_profile(RENDERING_MODE_NAME_STR, self._update_render_mode)

        self.zoom_factor_spin_box = CustomDoubleSpinBox()
        self.zoom_factor_spin_box.setValue(settings.get_from_profile(SEARCH_ZOOM_FACTOR_STR, 1.2))
        self.zoom_factor_spin_box.valueChanged.connect(self.change_zoom_factor)
        settings.connect_to_profile(SEARCH_ZOOM_FACTOR_STR, self._update_zoom_factor)

        self.scale_bar_ticks = QCheckBox()
        self.scale_bar_ticks.setChecked(settings.get_from_profile("scale_bar_ticks", True))
        self.scale_bar_ticks.stateChanged.connect(self.change_scale_bar_ticks)
        settings.connect_to_profile("scale_bar_ticks", self._update_scale_bar_ticks)

        layout = QGridLayout()
        layout.addWidget(QLabel("Theme:"), 0, 0)
        layout.addWidget(self.layout_list, 0, 1)
        layout.addWidget(QLabel("ROI render mode:"), 1, 0)
        layout.addWidget(self.labels_render_cmb, 1, 1)
        layout.addWidget(QLabel("Zoom factor for search ROI:"), 2, 0)
        layout.addWidget(self.zoom_factor_spin_box, 2, 1)
        layout.addWidget(QLabel("Show scale bar ticks"), 3, 0)
        layout.addWidget(self.scale_bar_ticks, 3, 1)
        layout.setColumnStretch(2, 1)
        layout.setRowStretch(6, 1)
        self.setLayout(layout)

    def change_scale_bar_ticks(self):
        self.settings.set_in_profile("scale_bar_ticks", self.scale_bar_ticks.isChecked())

    def _update_scale_bar_ticks(self):
        self.scale_bar_ticks.setChecked(self.settings.get_from_profile("scale_bar_ticks", True))

    def change_zoom_factor(self):
        self.settings.set_in_profile(SEARCH_ZOOM_FACTOR_STR, self.zoom_factor_spin_box.value())

    def _update_zoom_factor(self):
        self.zoom_factor_spin_box.setValue(self.settings.get_from_profile(SEARCH_ZOOM_FACTOR_STR))

    def change_theme(self):
        self.settings.theme_name = self.layout_list.currentText()

    def change_render_mode(self, text):
        self.settings.set_in_profile(RENDERING_MODE_NAME_STR, text)

    def _update_render_mode(self):
        self.labels_render_cmb.setCurrentText(
            self.settings.get_from_profile(RENDERING_MODE_NAME_STR, RENDERING_LIST[0])
        )


class ColorControl(QTabWidget):
    """
    Class for storage all settings for labels and colormaps.
    """

    def __init__(self, settings: BaseSettings, image_view_names: list[str]):
        super().__init__()
        self.appearance = Appearance(settings)
        self.colormap_editor = PColormapCreator(settings)
        self.color_preview = PColormapList(settings, image_view_names)
        self.color_preview.edit_signal.connect(self.colormap_editor.set_colormap)
        self.color_preview.edit_signal.connect(self._set_colormap_editor)
        self.label_editor = LabelEditor(settings)
        self.label_view = LabelChoose(settings)
        self.label_view.edit_with_name_signal.connect(self.label_editor.set_colors)
        self.label_view.edit_signal.connect(self._set_label_editor)
        self.mask_control = MaskControl(settings)
        self.addTab(self.appearance, "Appearance")
        self.addTab(self.color_preview, "Color maps")
        self.addTab(self.colormap_editor, "Color Map creator")
        self.addTab(self.label_view, "Select labels")
        self.addTab(self.label_editor, "Create labels")
        self.addTab(self.mask_control, "Mask marking")

    def _set_label_editor(self):
        self.setCurrentWidget(self.label_editor)

    def _set_colormap_editor(self):
        self.setCurrentWidget(self.colormap_editor)


class AdvancedWindow(QTabWidget):
    """
    Base class for advanced windows.
    It contains colormap management connected tabs: :py:class:`.PColormapCreator`
    and :py:class:`.PColormapList`.

    :param settings: program settings
    :param image_view_names: passed as second argument to :py:class:`~.PColormapList`
    """

    def __init__(self, settings: BaseSettings, image_view_names: list[str], reload_list=None, parent=None):
        super().__init__(parent)
        self.color_control = ColorControl(settings, image_view_names)
        self.settings = settings
        self.reload_list = reload_list if reload_list is not None else []

        self.develop = DevelopTab()
        self.addTab(self.color_control, "Color control")
        if state_store.develop:
            self.addTab(self.develop, "Develop")
        if self.window() == self:
            with suppress(KeyError):
                geometry = self.settings.get_from_profile("advanced_window_geometry")
                self.restoreGeometry(QByteArray.fromHex(bytes(geometry, "ascii")))

    def closeEvent(self, event: QCloseEvent) -> None:
        """
        Save geometry if widget is used as standalone window.
        """
        if self.window() == self:
            self.settings.set_in_profile("advanced_window_geometry", self.saveGeometry().toHex().data().decode("ascii"))
        super().closeEvent(event)


class ImageMetadata(QWidget):
    def __init__(self, settings: PartSettings, parent=None):
        super().__init__(parent)
        self.settings = settings
        self._dict_viewer = DictViewer(self.settings.image.metadata if self.settings.image else None)
        self.channel_info = QLabel()
        layout = QVBoxLayout()
        layout.addWidget(self._dict_viewer)
        layout.addWidget(self.channel_info)
        self.setLayout(layout)
        self.settings.image_changed.connect(self.update_metadata)

    def update_metadata(self):
        self._dict_viewer.set_data(self.settings.image.metadata)
        text = ", ".join(
            f"{name}: {color}"
            for name, color in zip(self.settings.image.channel_names, self.settings.image.get_colors())
        )
        self.channel_info.setText(f"Channels with colors: {text}")



import numpy as np
from scipy.spatial import ConvexHull

try:
    from scipy.spatial import QhullError
except ImportError:
    # Scipy below 1.8.0
    from scipy.spatial.qhull import QhullError
# this two functions are from
# https://stackoverflow.com/questions/37117878/generating-a-filled-polygon-inside-a-numpy-array/37123933#37123933


def check(p1, p2, idxs):
    """
    Uses the line defined by p1 and p2 to check array of
    input indices against interpolated value

    Returns boolean array, with True inside and False outside of shape
    """
    p1 = p1.astype(float)
    p2 = p2.astype(float)

    # Calculate max column idx for each row idx based on interpolated line between two points
    if p1[0] == p2[0]:
        max_col_idx = (idxs[0] - p1[0]) * idxs.shape[1]
        sign = np.sign(p2[1] - p1[1])
    else:
        max_col_idx = (idxs[0] - p1[0]) / (p2[0] - p1[0]) * (p2[1] - p1[1]) + p1[1]
        sign = np.sign(p2[0] - p1[0])
    return idxs[1] * sign <= max_col_idx * sign


def create_polygon(shape, vertices):
    """
    Creates np.array with dimensions defined by shape
    Fills polygon defined by vertices with ones, all other values zero"""
    idxs = np.indices(shape)  # Create 3D array of indices
    base_array = np.zeros(shape, dtype=float)  # Initialize your array of zeros

    fill = np.ones(base_array.shape) * True  # Initialize boolean array defining shape fill

    # Create check array for each edge segment, combine into fill array
    for k in range(vertices.shape[0]):
        fill = np.all([fill, check(vertices[k - 1], vertices[k], idxs)], axis=0)

    # Set all values inside polygon to one
    base_array[fill] = 1

    return base_array


def _convex_fill(array: np.ndarray):
    if array.ndim != 2:
        raise ValueError("Convex fill need to be called on 2d array.")
    points = np.transpose(np.nonzero(array))
    try:
        convex = ConvexHull(points)
        convex_points = points[convex.vertices]
        convex.close()
        return create_polygon(array.shape, convex_points[::-1])
    except (QhullError, ValueError):
        return None


def convex_fill(array: np.ndarray):
    arr_shape = array.shape
    array = np.squeeze(array)
    if array.ndim not in [2, 3]:
        raise ValueError("Convex hull support only 2 and 3 dimension images")
    components = np.bincount(array.flat)
    for i in range(1, components.size):
        if components[i] == 0:
            continue
        component: np.ndarray = array == i
        points = np.nonzero(component)
        if len(points) == 0 or len(points[0]) == 0:  # pylint: disable=consider-using-in
            continue
        lower_bound = np.min(points, axis=1)
        upper_bound = np.max(points, axis=1)
        cut_area = tuple(slice(x, y + 1) for x, y in zip(lower_bound, upper_bound))
        if array.ndim == 3:
            cut_area = (slice(None),) + cut_area[1:]
        component = component[cut_area]
        if array.ndim == 2:
            res = _convex_fill(component)
            if res is None:
                continue
            array[cut_area][res > 0] = i
        elif array.ndim == 3:
            for j in range(lower_bound[0], upper_bound[0] + 1):
                res = _convex_fill(component[j])
                if res is None:
                    continue
                new_cut = (j,) + cut_area[1:]
                tmp = array[new_cut]
                tmp[res > 0] = i
                array[new_cut] = tmp
    return array.reshape(arr_shape)



# pylint: disable=no-self-use

import os.path
from typing import TYPE_CHECKING

import numpy as np
import pytest

from PartSegCore.analysis.algorithm_description import AnalysisAlgorithmSelection
from PartSegCore.io_utils import load_metadata_base
from PartSegCore.utils import check_loaded_dict
from PartSegImage import TiffImageReader

if TYPE_CHECKING:
    from PartSegCore.algorithm_describe_base import ROIExtractionProfile
    from PartSegCore.segmentation.algorithm_base import ROIExtractionAlgorithm


def empty(_a, _b):
    pass  # pragma: no cover


class TestSegmentation:
    def test_profile_execute(self, data_test_dir):
        profile_path = os.path.join(data_test_dir, "segment_profile_test.json")
        # noinspection PyBroadException
        try:
            data = load_metadata_base(profile_path)
            assert check_loaded_dict(data)
        except Exception:  # pylint: disable=broad-except  # pragma: no cover
            pytest.fail("Fail in loading profile")
            return
        image = TiffImageReader.read_image(
            os.path.join(data_test_dir, "stack1_components", "stack1_component5.tif"),
            os.path.join(data_test_dir, "stack1_components", "stack1_component5_mask.tif"),
        )

        val: ROIExtractionProfile
        for val in data.values():
            algorithm: ROIExtractionAlgorithm = AnalysisAlgorithmSelection[val.algorithm]()
            algorithm.set_image(image)
            algorithm.set_mask(image.mask.squeeze())
            algorithm.set_parameters(val.values)
            result = algorithm.calculation_run(empty)
            assert np.max(result.roi) == 2



import contextlib
from contextlib import suppress

import pytest
from qtpy.QtWidgets import QDialog, QInputDialog, QMessageBox

from PartSeg._roi_analysis.partseg_settings import PartSettings
from PartSeg._roi_mask.main_window import ChosenComponents
from PartSeg._roi_mask.stack_settings import StackSettings
from PartSeg.common_backend.base_settings import BaseSettings
from PartSeg.common_gui import napari_image_view


@pytest.fixture
def base_settings(image, tmp_path, measurement_profiles):
    settings = BaseSettings(tmp_path)
    settings.image = image
    return settings


@pytest.fixture
def part_settings(image, tmp_path, measurement_profiles):
    settings = PartSettings(tmp_path)
    settings.image = image
    for el in measurement_profiles:
        settings.measurement_profiles[el.name] = el
    return settings


@pytest.fixture
def stack_settings(tmp_path, image):
    settings = StackSettings(tmp_path)
    settings.image = image
    chose = ChosenComponents()
    settings.chosen_components_widget = chose
    yield settings
    chose.deleteLater()


@pytest.fixture
def part_settings_with_project(image, analysis_segmentation2, tmp_path):
    settings = PartSettings(tmp_path)
    settings.image = image
    settings.set_project_info(analysis_segmentation2)
    return settings


@pytest.fixture(autouse=True)
def _disable_threads_viewer_patch_prepare_leyers(monkeypatch):
    def _prepare_layers(self, image, parameters, replace):
        self._add_image(napari_image_view._prepare_layers(image, parameters, replace))

    monkeypatch.setattr(napari_image_view.ImageView, "_prepare_layers", _prepare_layers)


@pytest.fixture(autouse=True)
def _disable_threads_viewer_patch_add_layer(monkeypatch, request):
    if "no_patch_add_layer" in request.keywords:
        return

    def _add_layer_util(self, index, layer, filters):
        if layer not in self.viewer.layers:
            self.viewer.add_layer(layer)

        monkeypatch.setattr(napari_image_view.ImageView, "_add_layer_util", _add_layer_util)


@pytest.fixture(autouse=True)
def _check_opened_windows(qapp):
    yield
    widgets = qapp.topLevelWidgets()
    for widget in widgets:
        assert not widget.isVisible()


@pytest.fixture(autouse=True)
def _block_threads(monkeypatch, request):
    if "enablethread" in request.keywords:
        return

    from pytestqt.qt_compat import qt_api
    from qtpy.QtCore import QThread, QTimer

    old_start = QTimer.start

    class OldTimer(QTimer):
        def start(self, time=None):
            if time is not None:
                old_start(self, time)
            else:
                old_start(self)

    def not_start(self):
        raise RuntimeError("Thread should not be used in test")

    monkeypatch.setattr(QTimer, "start", not_start)
    monkeypatch.setattr(QThread, "start", not_start)
    monkeypatch.setattr(qt_api.QtCore, "QTimer", OldTimer)


@pytest.fixture(autouse=True)
def _clean_settings():
    try:
        try:
            from napari.settings import SETTINGS
        except ImportError:
            from napari.utils.settings import SETTINGS
        SETTINGS.reset()
        yield
        with suppress(AttributeError):
            SETTINGS.reset()
    except ImportError:
        yield


@pytest.fixture(autouse=True)
def _reset_napari_settings(monkeypatch, tmp_path):
    def _mock_save(self, path=None, **dict_kwargs):
        return  # skipcq: PTC-W0049

    from napari import settings

    cp = settings.NapariSettings.__private_attributes__["_config_path"]
    monkeypatch.setattr(cp, "default", tmp_path / "save.yaml")
    monkeypatch.setattr(settings.NapariSettings, "save", _mock_save)
    settings._SETTINGS = None


@pytest.fixture(autouse=True)
def _block_message_box(monkeypatch, request):
    def raise_on_call(*_, **__):
        raise RuntimeError("exec_ call")  # pragma: no cover

    monkeypatch.setattr(QMessageBox, "exec_", raise_on_call)
    monkeypatch.setattr(QMessageBox, "critical", raise_on_call)
    monkeypatch.setattr(QMessageBox, "information", raise_on_call)
    monkeypatch.setattr(QMessageBox, "question", raise_on_call)
    monkeypatch.setattr(QMessageBox, "warning", raise_on_call)
    monkeypatch.setattr("PartSeg.common_gui.error_report.QMessageFromException.exec_", raise_on_call)
    monkeypatch.setattr(QInputDialog, "getText", raise_on_call)
    if "enabledialog" not in request.keywords:
        monkeypatch.setattr(QDialog, "exec_", raise_on_call)


class DummyConnect:
    def __init__(self, li):
        self.li = li

    def connect(self, func):
        self.li.append(func)


class DummyThrottler:
    def __init__(self, *args, **kwargs):
        self._call_list = []

    def setTimeout(self, *args, **kwargs):
        pass  # as it is dummy throttler then timeout is obsolete.

    def throttle(self, *args, **kwargs):
        for cl in self._call_list:
            cl(*args, **kwargs)

    @property
    def triggered(self):
        return DummyConnect(self._call_list)


@pytest.fixture(autouse=True)
def _mock_throttler(monkeypatch):
    with contextlib.suppress(ImportError):
        from napari._qt import qt_main_window

        if hasattr(qt_main_window, "QSignalThrottler"):
            monkeypatch.setattr(qt_main_window, "QSignalThrottler", DummyThrottler)
    with contextlib.suppress(ImportError):
        monkeypatch.setattr(
            "napari._qt.threads.status_checker.StatusChecker.start",
            lambda x: None,
            raising=False,
        )



from collections import defaultdict
from typing import Any, NamedTuple, Optional

import numpy as np

from PartSegCore.utils import numpy_repr
from PartSegCore_compiled_backend.utils import calc_bounds
from PartSegImage.image import Image, minimal_dtype


class BoundInfo(NamedTuple):
    """
    Information about bounding box
    """

    lower: np.ndarray
    upper: np.ndarray

    def box_size(self) -> np.ndarray:
        """Size of bounding box"""
        return self.upper - self.lower + 1

    def get_slices(self, margin=0) -> list[slice]:
        return [slice(max(x - margin, 0), y + 1 + margin) for x, y in zip(self.lower, self.upper)]

    def del_dim(self, axis: int):
        return BoundInfo(np.delete(self.lower, axis), np.delete(self.upper, axis))

    def __str__(self):
        return f"{self.__class__.__name__}(lower={list(self.lower)}, upper={list(self.upper)})"


class ROIInfo:
    """
    Object to storage meta information about given segmentation.
    Segmentation array is only referenced, not copied.

    :ivar numpy.ndarray ~.roi: reference to segmentation
    :ivar Dict[int,BoundInfo] bound_info: mapping from component number to bounding box
    :ivar numpy.ndarray sizes: array with sizes of components
    :ivar Dict[int, Any] annotations: annotations of roi
    :ivar Dict[str, np.ndarray] alternative: alternative representation of roi
    """

    def __init__(
        self,
        roi: Optional[np.ndarray],
        annotations: Optional[dict[int, Any]] = None,
        alternative: Optional[dict[str, np.ndarray]] = None,
    ):
        annotations = {} if annotations is None else annotations
        self.annotations = {int(k): v for k, v in annotations.items()}
        self.alternative = {} if alternative is None else alternative
        if roi is None:
            self.roi = None
            self.bound_info = {}
            self.sizes = []
            return
        max_val = np.max(roi)
        dtype = minimal_dtype(max_val)
        roi = roi.astype(dtype)
        self.roi = roi
        self.bound_info = self.calc_bounds(roi)
        self.sizes = np.bincount(roi.flat)
        self._alternative_component_size = {}

    def fit_to_image(self, image: Image) -> "ROIInfo":
        if self.roi is None:
            return ROIInfo(self.roi, self.annotations, self.alternative)
        roi = image.fit_array_to_image(self.roi)
        alternatives = {k: image.fit_array_to_image(v) for k, v in self.alternative.items()}
        return ROIInfo(roi, self.annotations, alternatives)

    def get_components_num(self, name):
        """
        Get the number of components for a given representation.

        Args:
            name (str): Name of the representation. Use "ROI" for the main ROI,
                       or the name of an alternative representation.

        Returns:
            int: Maximum component number in the specified representation.
        """
        if name == "ROI" or name not in self.alternative:
            return max(self.bound_info)
        if name not in self._alternative_component_size:
            self._alternative_component_size[name] = np.max(self.alternative[name])
        return self._alternative_component_size[name]

    def __str__(self):
        return f"ROIInfo; components: {len(self.bound_info)}, sizes: {self.sizes}"

    def __repr__(self):
        return f"ROIInfo(roi={numpy_repr(self.roi)}, bound_info={self.bound_info}, sizes={self.sizes!r})"

    @staticmethod
    def calc_bounds(roi: np.ndarray) -> dict[int, BoundInfo]:
        """
        Calculate bounding boxes components

        :param np.ndarray roi: array for which bounds boxes should be calculated
        :return: mapping component number to bounding box
        :rtype: Dict[int, BoundInfo]
        """
        try:
            min_bounds, max_bounds = calc_bounds(roi)
            return {
                num: BoundInfo(lower=lower, upper=upper)
                for num, (lower, upper) in enumerate(zip(min_bounds, max_bounds))
                if num != 0 and upper[0] != -1
            }
        except KeyError:
            bound_info = {}
            points = np.nonzero(roi)
            comp_num = roi[points]
            point_dict = defaultdict(list)
            for num, point in zip(comp_num, np.transpose(points)):
                point_dict[num].append(point)
            for num, points_for_num in point_dict.items():
                lower = np.min(points_for_num, 0)
                upper = np.max(points_for_num, 0)
                bound_info[num] = BoundInfo(lower=lower, upper=upper)
            return bound_info



from PartSeg.plugins.old_partseg.old_partseg import LoadPartSegOld


def register():
    from PartSeg import state_store

    if state_store.custom_plugin_load:
        from PartSegCore.register import RegisterEnum
        from PartSegCore.register import register as register_fun

        register_fun(LoadPartSegOld, RegisterEnum.analysis_load)



import tarfile
import time
import zipfile
from unittest.mock import MagicMock, patch

import pandas as pd
import pytest

from PartSeg._roi_analysis.export_batch import (
    MISSING_FILES,
    NO_FILES,
    ExportProjectDialog,
    ZenodoCreateError,
    _extract_information_from_excel_to_export,
    export_to_archive,
    export_to_zenodo,
    sleep_with_rate,
)


@pytest.fixture
def _dummy_tiffs(tmp_path):
    for i in range(1, 11):
        (tmp_path / f"stack1_component{i}.tif").touch()


def test_extract_information_from_excel_to_export_no_file(tmp_path, bundle_test_dir):
    res = _extract_information_from_excel_to_export(bundle_test_dir / "sample_batch_output.xlsx", tmp_path)
    assert all(not x[1] for x in res)


def test_extract_information_from_excel_to_export(data_test_dir, bundle_test_dir):
    res = _extract_information_from_excel_to_export(
        bundle_test_dir / "sample_batch_output.xlsx", data_test_dir / "stack1_components"
    )
    assert all(x[1] for x in res)


def all_files_in_archive(path):
    if path.suffix == ".zip":
        with zipfile.ZipFile(path) as zip_file:
            name_list = zip_file.namelist()
    else:
        with tarfile.open(path) as tar_file:
            name_list = tar_file.getnames()

    assert set(name_list) == {"sample_batch_output.xlsx", *[f"stack1_component{i}.tif" for i in range(1, 9)]}


@pytest.mark.usefixtures("_dummy_tiffs")
@pytest.mark.parametrize("ext", [".zip", ".tar.gz", ".txz", ".tar.bz2"])
def test_export_to_archive(bundle_test_dir, tmp_path, ext):
    a = list(export_to_archive(bundle_test_dir / "sample_batch_output.xlsx", tmp_path, tmp_path / f"arch{ext}"))
    assert len(a) == 9
    all_files_in_archive(tmp_path / f"arch{ext}")


@pytest.mark.usefixtures("_dummy_tiffs")
def test_fail_export_to_archive_unknown_extension(bundle_test_dir, tmp_path):
    with pytest.raises(ValueError, match="Unknown archive type"):
        list(export_to_archive(bundle_test_dir / "sample_batch_output.xlsx", tmp_path, tmp_path / "arch.bla"))


@pytest.mark.usefixtures("_dummy_tiffs")
def test_fail_export_to_archive_missing_file(bundle_test_dir, tmp_path):
    (tmp_path / "stack1_component1.tif").unlink()
    with pytest.raises(ValueError, match=MISSING_FILES):
        list(export_to_archive(bundle_test_dir / "sample_batch_output.xlsx", tmp_path, tmp_path / "arch.tar.gz"))


def test_fail_export_empty_excel(tmp_path):
    pd.DataFrame().to_excel(tmp_path / "empty.xlsx")
    with pytest.raises(ValueError, match=NO_FILES):
        list(export_to_archive(tmp_path / "empty.xlsx", tmp_path, tmp_path / "arch.tar.gz"))


@pytest.fixture
def zenodo_kwargs():
    return {
        "zenodo_token": "sample_token",
        "title": "sample title",
        "author": "sample author",
        "affiliation": "sample affiliation",
        "orcid": "0000-0000-0000-0000",
        "description": "sample description",
        "zenodo_url": "https://dummy_url",
    }


@pytest.mark.usefixtures("_dummy_tiffs")
@patch("requests.post")
@patch("requests.put")
def test_zenodo_export(put_mock, post_mock, bundle_test_dir, tmp_path, zenodo_kwargs):
    post_mock.return_value.status_code = 201
    put_mock.return_value.status_code = 200
    a = list(
        export_to_zenodo(excel_path=bundle_test_dir / "sample_batch_output.xlsx", base_folder=tmp_path, **zenodo_kwargs)
    )
    assert len(a) == 9
    assert post_mock.call_count == 1
    assert put_mock.call_count == 10


@pytest.mark.usefixtures("_dummy_tiffs")
@patch("requests.post")
@patch("requests.put")
def test_zenodo_export_fail_no_files(put_mock, post_mock, bundle_test_dir, tmp_path, zenodo_kwargs):
    (tmp_path / "stack1_component1.tif").unlink()
    with pytest.raises(ValueError, match=MISSING_FILES):
        list(
            export_to_zenodo(
                excel_path=bundle_test_dir / "sample_batch_output.xlsx", base_folder=tmp_path, **zenodo_kwargs
            )
        )
    assert post_mock.call_count == 0
    assert put_mock.call_count == 0


@pytest.mark.usefixtures("_dummy_tiffs")
@patch("requests.post")
@patch("requests.put")
def test_zenodo_export_fail(put_mock, post_mock, tmp_path, zenodo_kwargs):
    pd.DataFrame().to_excel(tmp_path / "empty.xlsx")
    with pytest.raises(ValueError, match=NO_FILES):
        list(export_to_zenodo(excel_path=tmp_path / "empty.xlsx", base_folder=tmp_path, **zenodo_kwargs))
    assert post_mock.call_count == 0
    assert put_mock.call_count == 0


@pytest.mark.usefixtures("_dummy_tiffs")
@patch("requests.post")
@patch("requests.put")
def test_zenodo_export_fail_create_deposit(put_mock, post_mock, bundle_test_dir, tmp_path, zenodo_kwargs):
    post_mock.return_value.status_code = 400
    with pytest.raises(ZenodoCreateError, match="Can't create deposition"):
        list(
            export_to_zenodo(
                excel_path=bundle_test_dir / "sample_batch_output.xlsx", base_folder=tmp_path, **zenodo_kwargs
            )
        )
    assert post_mock.call_count == 1
    assert put_mock.call_count == 0


@patch("time.sleep")
def test_sleep_with_rate(sleep_mock):
    response_mock = MagicMock()
    sleep_with_rate(response_mock)
    sleep_mock.assert_not_called()
    response_mock.headers = {"X-RateLimit-Remaining": 10}
    sleep_with_rate(response_mock)
    sleep_mock.assert_not_called()
    response_mock.headers["X-RateLimit-Remaining"] = 0
    response_mock.headers["X-RateLimit-Reset"] = time.time() + 100
    sleep_with_rate(response_mock)
    sleep_mock.assert_called_once()


class TestExportProjectDialog:
    def test_create(self, part_settings, qtbot):
        dlg = ExportProjectDialog("", "", part_settings)
        qtbot.addWidget(dlg)
        assert not dlg.export_btn.isEnabled()
        assert not dlg.export_to_zenodo_btn.isEnabled()

    @pytest.mark.parametrize("ext", ["", ".xlsx"])
    def test_set_excel_file(self, monkeypatch, part_settings, qtbot, bundle_test_dir, ext):
        dlg = ExportProjectDialog("", "", part_settings)
        qtbot.addWidget(dlg)
        monkeypatch.setattr("PartSeg.common_gui.custom_load_dialog.PLoadDialog.exec_", lambda x: True)
        monkeypatch.setattr(
            "PartSeg.common_gui.custom_load_dialog.PLoadDialog.selectedFiles",
            lambda x: [str(bundle_test_dir / ("sample_batch_output" + ext))],
        )
        assert dlg.excel_path.text() == ""
        dlg.select_excel()
        assert dlg.excel_path.text() == str(bundle_test_dir / "sample_batch_output.xlsx")

    def test_set_base_folder(self, monkeypatch, part_settings, qtbot, bundle_test_dir):
        dlg = ExportProjectDialog("", "", part_settings)
        qtbot.addWidget(dlg)
        monkeypatch.setattr("PartSeg.common_gui.custom_load_dialog.SelectDirectoryDialog.exec_", lambda x: True)
        monkeypatch.setattr(
            "PartSeg.common_gui.custom_load_dialog.SelectDirectoryDialog.selectedFiles",
            lambda x: [str(bundle_test_dir)],
        )
        assert dlg.base_folder.text() == ""
        dlg.select_folder()
        assert dlg.base_folder.text() == str(bundle_test_dir)

    @pytest.mark.usefixtures("_dummy_tiffs")
    def test_enable_export_btn(self, part_settings, qtbot, bundle_test_dir, tmp_path):
        dlg = ExportProjectDialog("", "", part_settings)
        qtbot.addWidget(dlg)
        assert not dlg.export_btn.isEnabled()

        dlg.excel_path.setText(str(bundle_test_dir / "sample_batch_output.xlsx"))
        assert not dlg.export_btn.isEnabled()

        assert dlg.info_box.topLevelItemCount() == 8

        dlg.base_folder.setText(str(tmp_path))
        assert dlg.export_btn.isEnabled()

        pd.DataFrame().to_excel(tmp_path / "empty.xlsx")
        dlg.excel_path.setText(str(tmp_path / "empty.xlsx"))
        assert not dlg.export_btn.isEnabled()
        assert dlg.info_label.text() == NO_FILES

    @pytest.mark.usefixtures("_dummy_tiffs")
    @patch("PartSeg._roi_analysis.export_batch.PSaveDialog")
    def test_export_to_zip(self, save_dlg_mock, part_settings, qtbot, bundle_test_dir, tmp_path, monkeypatch):
        dlg = ExportProjectDialog("", "", part_settings)
        qtbot.addWidget(dlg)
        assert not dlg.export_to_zenodo_btn.isEnabled()

        dlg.excel_path.setText(str(bundle_test_dir / "sample_batch_output.xlsx"))
        dlg.base_folder.setText(str(tmp_path))
        save_dlg_mock.exec_ = MagicMock(return_value=True)
        save_dlg_mock.selectedFiles = MagicMock(return_value=[str(tmp_path / "test.zip")])

        dlg.export_btn.click()
        qtbot.wait_until(lambda: dlg.worker is None)
        (tmp_path / "test.zip").exists()

    @pytest.mark.usefixtures("_dummy_tiffs")
    def test_enable_zenodo_export_btn(self, part_settings, qtbot, bundle_test_dir, tmp_path):
        dlg = ExportProjectDialog("", "", part_settings)
        qtbot.addWidget(dlg)
        assert not dlg.export_to_zenodo_btn.isEnabled()

        dlg.excel_path.setText(str(bundle_test_dir / "sample_batch_output.xlsx"))
        dlg.base_folder.setText(str(tmp_path))
        assert not dlg.export_to_zenodo_btn.isEnabled()

        dlg.zenodo_token.setText("test111")
        dlg.zenodo_author.setText("test")
        dlg.zenodo_affiliation.setText("test")
        dlg.zenodo_title.setText("test")
        assert not dlg.export_to_zenodo_btn.isEnabled()
        dlg.zenodo_description.setText("test")
        assert dlg.export_to_zenodo_btn.isEnabled()

    @pytest.mark.usefixtures("_dummy_tiffs")
    @pytest.mark.parametrize(
        ("sandbox", "zenodo_url"), [(True, "https://sandbox.zenodo.org"), (False, "https://zenodo.org")]
    )
    @patch("requests.post")
    @patch("requests.put")
    def test_export_to_zenodo(
        self, put_mock, post_mock, part_settings, qtbot, bundle_test_dir, tmp_path, sandbox, zenodo_url
    ):
        post_mock.return_value.status_code = 201
        put_mock.return_value.status_code = 200
        dlg = ExportProjectDialog("", "", part_settings)
        qtbot.addWidget(dlg)
        dlg.excel_path.setText(str(bundle_test_dir / "sample_batch_output.xlsx"))
        dlg.base_folder.setText(str(tmp_path))
        dlg.zenodo_token.setText("test111")
        dlg.zenodo_author.setText("test1")
        dlg.zenodo_affiliation.setText("test2")
        dlg.zenodo_title.setText("test3")
        dlg.zenodo_description.setText("test4")
        dlg.zenodo_orcid.setText("0000-0000-0000-0001")
        dlg.zenodo_sandbox.setChecked(sandbox)

        dlg.export_to_zenodo_btn.click()
        qtbot.wait_until(lambda: dlg.worker is None)
        assert post_mock.call_count == 1
        assert post_mock.call_args[0][0].startswith(zenodo_url)
        assert put_mock.call_count == 10
        assert part_settings.get("zenodo_token") == "test111"
        assert part_settings.get("zenodo_author") == "test1"
        assert part_settings.get("zenodo_affiliation") == "test2"
        assert part_settings.get("zenodo_orcid") == "0000-0000-0000-0001"



import itertools
import logging
import platform
from collections.abc import MutableMapping
from contextlib import suppress
from dataclasses import dataclass, field
from enum import Enum
from functools import partial
from typing import TYPE_CHECKING, Optional, Union

import napari
import numpy as np
from local_migrator import register_class
from napari.components import ViewerModel as Viewer
from napari.layers import Layer, Points
from napari.layers.image import Image as NapariImage
from napari.layers.labels import Labels
from napari.qt import QtViewer
from napari.qt.threading import thread_worker
from napari.utils.colormaps.colormap import ColormapInterpolationMode
from packaging.version import parse as parse_version
from qtpy.QtCore import QEvent, QPoint, Qt, QTimer, Signal, Slot
from qtpy.QtWidgets import QApplication, QCheckBox, QHBoxLayout, QLabel, QMenu, QSpinBox, QToolTip, QVBoxLayout, QWidget
from scipy.ndimage import binary_dilation
from superqt import QEnumComboBox, ensure_main_thread
from vispy.color import Color, Colormap
from vispy.geometry.rect import Rect

from PartSeg.common_backend.base_settings import BaseSettings
from PartSeg.common_gui.advanced_tabs import RENDERING_LIST, RENDERING_MODE_NAME_STR, SEARCH_ZOOM_FACTOR_STR
from PartSeg.common_gui.channel_control import ChannelProperty, ColorComboBoxGroup
from PartSeg.common_gui.custom_buttons import SearchROIButton
from PartSeg.common_gui.qt_modal import QtPopup
from PartSegCore.image_operations import NoiseFilterType, bilateral, gaussian, median
from PartSegCore.roi_info import ROIInfo
from PartSegImage import Image

if TYPE_CHECKING:
    from vispy.scene import BaseCamera

try:
    from napari._qt.qt_viewer_buttons import QtViewerPushButton as QtViewerPushButton_
except ImportError:
    from napari._qt.widgets.qt_viewer_buttons import QtViewerPushButton as QtViewerPushButton_
_napari_ge_4_13 = parse_version(napari.__version__) >= parse_version("0.4.13a1")
_napari_ge_4_17 = parse_version(napari.__version__) >= parse_version("0.4.17a1")
_napari_ge_4_19 = parse_version(napari.__version__) >= parse_version("0.4.19")
_napari_ge_5 = parse_version(napari.__version__) >= parse_version("0.5.0a1")

# if run with numpy<2 on macOS arm64 architecture compiled from pypi wheels
# then it will crash with bus error if numpy is used in different thread
if (
    parse_version(np.__version__) < parse_version("2")
    and platform.system() == "Darwin"
    and platform.machine() == "arm64"
):  # pragma: no cover
    try:
        USE_THREADS = "cibw-run" not in np.show_config("dicts")["Python Information"]["path"]
    except (KeyError, TypeError):
        USE_THREADS = True
else:
    USE_THREADS = True


class _NapariImage(NapariImage):
    def _update_thumbnail(self, *_, **__):
        """Disable thumbnail update"""


def get_highlight_colormap():
    cmap_dict = {0: (0, 0, 0, 0), 1: "white", None: (0, 0, 0, 0)}
    if _napari_ge_5:
        from napari.utils.colormaps import DirectLabelColormap

        return {"colormap": DirectLabelColormap(color_dict=cmap_dict)}

    return {"color": cmap_dict}


class QtViewerPushButton(QtViewerPushButton_):
    def __init__(self, viewer, *args, **kwargs):
        if _napari_ge_4_13:
            super().__init__(*args, **kwargs)
        else:
            super().__init__(viewer, *args, **kwargs)


if _napari_ge_4_13:

    class QtNDisplayButton(QtViewerPushButton_):
        def __init__(self, viewer):
            super().__init__(button_name="ndisplay_button", tooltip="Toggle dimensions", slot=self.toggle_ndisplay)
            self.viewer = viewer
            self.setCheckable(True)

        def toggle_ndisplay(self):
            self.viewer.dims.ndisplay = 2 + (self.viewer.dims.ndisplay == 2)

else:
    from napari.qt import QtStateButton

    class QtNDisplayButton(QtStateButton):
        def __init__(self, viewer):
            super().__init__(
                "ndisplay_button",
                viewer.dims,
                "ndisplay",
                viewer.dims.events.ndisplay,
                2,
                3,
            )


ORDER_DICT = {"xy": [0, 1, 2, 3], "zy": [0, 2, 1, 3], "zx": [0, 3, 1, 2]}
NEXT_ORDER = {"xy": "zy", "zy": "zx", "zx": "xy"}

ColorInfo = dict[Optional[int], Union[str, list[float]]]


@dataclass
class ImageInfo:
    image: Image
    layers: list[NapariImage]
    filter_info: list[tuple[NoiseFilterType, float]] = field(default_factory=list)
    mask: Optional[Labels] = None
    mask_array: Optional[np.ndarray] = None
    roi: Optional[Labels] = None
    roi_info: ROIInfo = field(default_factory=lambda: ROIInfo(None))
    roi_count: int = 0
    highlight: Optional[Labels] = None

    def coords_in(self, coords: Union[list[int], np.ndarray]) -> bool:
        if not self.layers:
            return False
        fst_layer = self.layers[0]
        moved_coords = self.translated_coords(coords)
        return np.all(moved_coords >= 0) and np.all(moved_coords < fst_layer.data.shape)

    def translated_coords(self, coords: Union[list[int], np.ndarray]) -> np.ndarray:
        if not self.layers:
            return np.array(coords)
        fst_layer = self.layers[0]
        return np.subtract(coords, fst_layer.translate).astype(int)


@register_class(old_paths=["PartSeg.common_gui.stack_image_view.LabelEnum"])
class LabelEnum(Enum):
    Not_show = 0
    Show_results = 1
    Show_selected = 2

    def __str__(self):
        return "Don't show" if self.value == 0 else self.name.replace("_", " ")


class SearchType(Enum):
    Highlight = 0
    Zoom_in = 1


class ImageView(QWidget):
    position_changed = Signal([int, int, int], [int, int])
    component_clicked = Signal(int)
    text_info_change = Signal(str)
    hide_signal = Signal(bool)
    view_changed = Signal()
    image_added = Signal()

    def __init__(
        self,
        settings: BaseSettings,
        channel_property: ChannelProperty,
        name: str,
        parent: Optional[QWidget] = None,
        ndisplay=2,
    ):
        super().__init__(parent=parent)

        self.settings = settings
        self.channel_property = channel_property
        self.name = name
        self.image_info: dict[str, ImageInfo] = {}
        self.current_image = ""
        self._current_order = "xy"
        self.components = None
        self.worker_list = []
        self.points_layer = None
        self.roi_alternative_selection = "ROI"
        self._search_type = SearchType.Highlight
        self._last_component = 1

        self.viewer = Viewer(ndisplay=ndisplay)
        self.viewer.theme = self.settings.theme_name
        self.viewer_widget = NapariQtViewer(self.viewer)

        self._update_scale_bar_ticks()

        self.channel_control = ColorComboBoxGroup(settings, name, channel_property, height=30)
        self.ndim_btn = QtNDisplayButton(self.viewer)
        self.reset_view_button = QtViewerPushButton(self.viewer, "home", "Reset view", self._reset_view)
        self.points_view_button = QtViewerPushButton(
            self.viewer, "new_points", "Show points", self.toggle_points_visibility
        )
        self.points_view_button.setVisible(False)
        self.search_roi_btn = SearchROIButton(self.settings)
        self.search_roi_btn.clicked.connect(self._search_component)
        self.search_roi_btn.setDisabled(True)
        self.roll_dim_button = QtViewerPushButton(self.viewer, "roll", "Roll dimension", self._rotate_dim)
        self.roll_dim_button.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.roll_dim_button.customContextMenuRequested.connect(self._dim_order_menu)
        self.mask_chk = QCheckBox()
        self.mask_chk.setVisible(False)
        self.mask_label = QLabel("Mask:")
        self.mask_label.setVisible(False)

        self.btn_layout = QHBoxLayout()
        self.btn_layout2 = QHBoxLayout()
        self.setup_ui()

        self.channel_control.change_channel.connect(self.change_visibility)
        self.viewer.events.status.connect(self.print_info)

        self._connect_to_settings()

        self.old_scene: BaseCamera = self.viewer_widget.view.scene

        self.mask_chk.stateChanged.connect(self.change_mask_visibility)
        self.viewer_widget.view.scene.transform.changed.connect(self._view_changed, position="last")
        self.viewer.dims.events.current_step.connect(self._view_changed, position="last")
        self.viewer.dims.events.ndisplay.connect(self._view_changed, position="last")
        self.viewer.dims.events.ndisplay.connect(self._view_changed, position="last")
        self.viewer.dims.events.ndisplay.connect(self.camera_change, position="last")
        self.viewer.events.reset_view.connect(self._view_changed, position="last")

    def setup_ui(self):
        self.btn_layout.addWidget(self.reset_view_button)
        self.btn_layout.addWidget(self.ndim_btn)
        self.btn_layout.addWidget(self.roll_dim_button)
        self.btn_layout.addWidget(self.points_view_button)
        self.btn_layout.addWidget(self.search_roi_btn)
        self.btn_layout.addWidget(self.channel_control, 1)
        self.btn_layout.addWidget(self.mask_label)
        self.btn_layout.addWidget(self.mask_chk)
        layout = QVBoxLayout()
        layout.addLayout(self.btn_layout)
        layout.addLayout(self.btn_layout2)
        layout.addWidget(self.viewer_widget)

        self.setLayout(layout)

        if hasattr(self.viewer_widget.canvas, "background_color_override"):
            self.viewer_widget.canvas.background_color_override = "black"
            if _napari_ge_4_17:
                self.viewer.scale_bar.color = "white"
                self.viewer.scale_bar.colored = True

    def _connect_to_settings(self):
        self.settings.mask_changed.connect(self.set_mask)
        self.settings.roi_changed.connect(self.set_roi)
        self.settings.roi_clean.connect(self.set_roi)
        self.settings.image_changed.connect(self.set_image)
        self.settings.image_spacing_changed.connect(self.update_spacing_info)
        self.settings.points_changed.connect(self.update_points)
        self.settings.connect_to_profile(RENDERING_MODE_NAME_STR, self.update_rendering)
        self.settings.labels_changed.connect(self.update_roi_coloring)
        self.settings.connect_to_profile(f"{self.name}.image_state.opacity", self.update_roi_coloring)
        self.settings.connect_to_profile(f"{self.name}.image_state.only_border", self.update_roi_border)
        self.settings.connect_to_profile(f"{self.name}.image_state.border_thick", self.update_roi_border)
        self.settings.connect_to_profile(f"{self.name}.image_state.show_label", self.update_roi_labeling)
        self.settings.connect_to_profile("mask_presentation_opacity", self.update_mask_parameters)
        self.settings.connect_to_profile("mask_presentation_color", self.update_mask_parameters)
        self.settings.connect_to_profile("scale_bar_ticks", self._update_scale_bar_ticks)

    def _update_scale_bar_ticks(self):
        self.viewer.scale_bar.ticks = self.settings.get_from_profile("scale_bar_ticks", True)

    def toggle_points_visibility(self):
        if self.points_layer is not None:
            self.points_layer.visible = not self.points_layer.visible

    def toggle_scale_bar(self):
        self.viewer.scale_bar.unit = "nm"
        self.viewer.scale_bar.visible = not self.viewer.scale_bar.visible

    def _dim_order_menu(self, point: QPoint):
        menu = QMenu()
        for key in ORDER_DICT:
            action = menu.addAction(key)
            action.triggered.connect(partial(self._set_new_order, key))
            if key == self._current_order:
                font = action.font()
                font.setBold(True)
                action.setFont(font)

        menu.exec_(self.roll_dim_button.mapToGlobal(point))

    def _set_new_order(self, text: str):
        self._current_order = text
        self.viewer.dims.order = ORDER_DICT[text]

    def _reset_view(self):
        self._set_new_order("xy")
        self.viewer.dims.order = ORDER_DICT[self._current_order]
        self.viewer.reset_view()

    def _rotate_dim(self):
        self._set_new_order(NEXT_ORDER[self._current_order])

    def camera_change(self, _args):
        self.old_scene.transform.changed.disconnect(self._view_changed)
        self.old_scene: BaseCamera = self.viewer_widget.view.camera
        self.old_scene.transform.changed.connect(self._view_changed, position="last")

    def _view_changed(self, _args):
        self.view_changed.emit()

    def get_state(self):
        return {
            "ndisplay": self.viewer.dims.ndisplay,
            "point": self.viewer.dims.point,
            "camera": self.viewer_widget.view.camera.get_state(),
        }

    def set_state(self, dkt):
        if "ndisplay" in dkt and self.viewer.dims.ndisplay != dkt["ndisplay"]:
            self.viewer.dims.ndisplay = dkt["ndisplay"]
            return
        if "point" in dkt:
            for i, val in enumerate(dkt["point"]):
                self.viewer.dims.set_point(i, val)
        if "camera" in dkt:
            with suppress(KeyError):
                self.viewer_widget.view.camera.set_state(dkt["camera"])

    def change_mask_visibility(self):
        for image_info in self.image_info.values():
            if image_info.mask is not None:
                image_info.mask.visible = self.mask_chk.isChecked()

    def update_spacing_info(self, image: Optional[Image] = None) -> None:
        """
        Update spacing of image if not provide, then use image pointed by settings.

        :param Optional[Image] image: image which spacing should be updated.
        :return: None
        """
        if image is None:
            image = self.settings.image

        if image.file_path not in self.image_info:
            raise ValueError("Image not registered")

        image_info = self.image_info[image.file_path]

        for layer in image_info.layers:
            layer.scale = image.normalized_scaling()

        if image_info.roi is not None:
            image_info.roi.scale = image.normalized_scaling()

        if image_info.mask is not None:
            image_info.mask.scale = image.normalized_scaling()

    def _active_layer(self):
        if hasattr(self.viewer.layers, "selection"):
            return self.viewer.layers.selection.active
        return self.viewer.active_layer

    def _coordinates(self):
        active_layer = self._active_layer()
        if active_layer is None:
            return None
        if (
            hasattr(self.viewer, "cursor")
            and hasattr(self.viewer.cursor, "position")
            and hasattr(active_layer, "world_to_data")
        ):
            return [int(x) for x in active_layer.world_to_data(self.viewer.cursor.position)]
        return [int(x) for x in active_layer.coordinates]

    def print_info(self, event=None):
        cords = self._coordinates()
        if cords is None:
            return
        bright_array = []
        components = []
        alt_components = []
        for image_info in self.image_info.values():
            if not image_info.coords_in(cords):
                continue
            moved_coords = image_info.translated_coords(cords)
            bright_array.extend(layer.data[tuple(moved_coords)] for layer in image_info.layers if layer.visible)

            if (
                image_info.roi_info.roi is not None
                and image_info.roi is not None
                and (val := image_info.roi_info.roi[tuple(moved_coords)])
            ):
                components.append(val)
            if self.roi_alternative_selection in image_info.roi_info.alternative and (
                val := image_info.roi_info.alternative[self.roi_alternative_selection][tuple(moved_coords)]
            ):
                alt_components.append(val)

        if not bright_array and not components:
            self.text_info_change.emit("")
            return
        text = f"{cords}: "
        if bright_array:
            text += str(bright_array[0]) if len(bright_array) == 1 else "[" + ", ".join(map(str, bright_array)) + "]"
        self.components = components
        text += _print_list(components, " component")
        text += _print_list(alt_components, " alt")
        self.text_info_change.emit(text)

    def mask_opacity(self) -> float:
        """Get mask opacity"""
        return self.settings.get_from_profile("mask_presentation_opacity", 1)

    def mask_color(self) -> ColorInfo:
        """Get mask marking color"""
        color = Color(np.divide(self.settings.get_from_profile("mask_presentation_color", [255, 255, 255]), 255))
        return {0: (0, 0, 0, 0), 1: color.rgba, None: (0, 0, 0, 0)}

    def get_image(self, image: Optional[Image]) -> Image:
        if image is not None:
            return image
        if self.current_image not in self.image_info:
            return self.settings.image
        return self.image_info[self.current_image].image

    def update_points(self):
        if self.settings.points is not None:
            self.points_view_button.setVisible(True)
            if self.points_layer is None or self.points_layer not in self.viewer.layers:
                self.points_layer = Points(self.settings.points, scale=self.settings.image.normalized_scaling())
                self.viewer.add_layer(self.points_layer)
            else:
                self.points_layer.data = self.settings.points
                self.points_layer.scale = self.settings.image.normalized_scaling()
        elif self.points_layer is not None and self.points_layer in self.viewer.layers:
            self.points_view_button.setVisible(False)
            self.points_layer.data = np.empty((0, 4))

    def set_roi(self, roi_info: Optional[ROIInfo] = None, image: Optional[Image] = None) -> None:
        image = self.get_image(image)
        if roi_info is None:
            roi_info = self.settings.roi_info
        if image.file_path not in self.image_info:
            return
        image_info = self.image_info[image.file_path]
        if image_info.roi is None and roi_info.roi is not None:
            image_info.roi_info = roi_info
            self.add_roi_layer(image_info)
            self.search_roi_btn.setDisabled(False)
        elif image_info.roi is None:
            return
        elif roi_info.roi is None:
            image_info.roi.visible = False
            self.search_roi_btn.setDisabled(True)
            return
        else:
            image_info.roi_info = roi_info
            image_info.roi.data = roi_info.alternative.get(self.roi_alternative_selection, roi_info.roi)
            image_info.roi.visible = True
            self.search_roi_btn.setDisabled(False)

        image_info.roi_count = max(roi_info.bound_info) if roi_info.bound_info else 0

        if image_info.roi not in self.viewer.layers:
            self.viewer.add_layer(image_info.roi)

        self.set_roi_colormap(image_info)
        image_info.roi.opacity = self.settings.get_from_profile(f"{self.name}.image_state.opacity", 1.0)
        image_info.roi.refresh()

    def get_roi_view_parameters(self, image_info: ImageInfo) -> ColorInfo:
        colors = self.settings.label_colors / 255
        if (
            self.settings.get_from_profile(f"{self.name}.image_state.show_label", LabelEnum.Show_results)
            == LabelEnum.Not_show
            or image_info.roi_count == 0
            or colors.size == 0
        ):
            return {0: [0, 0, 0, 0], None: [0, 0, 0, 0]}
        res = {
            x: colors[(x - 1) % colors.shape[0]]
            for x in range(1, image_info.roi_info.get_components_num(self.roi_alternative_selection) + 1)
        }
        res[0] = [0, 0, 0, 0]
        res[None] = [0, 0, 0, 0]
        return res

    def set_roi_colormap(self, image_info) -> None:
        if _napari_ge_4_19:
            from napari.utils.colormaps import DirectLabelColormap

            image_info.roi.colormap = DirectLabelColormap(color_dict=self.get_roi_view_parameters(image_info))
            return
        if _napari_ge_4_13:
            image_info.roi.color = self.get_roi_view_parameters(image_info)
            return
        colors = self.settings.label_colors / 255
        if (
            self.settings.get_from_profile(f"{self.name}.image_state.show_label", LabelEnum.Show_results)
            == LabelEnum.Not_show
            or image_info.roi_count == 0
            or colors.size == 0
        ):
            image_info.roi.colormap = Colormap([[0, 0, 0, 0], [0, 0, 0, 0]])

        res = [[*list(colors[(x - 1) % colors.shape[0]]), 1] for x in range(image_info.roi_count + 1)]
        res[0] = [0, 0, 0, 0]
        if len(res) < 2:
            res += [[0, 0, 0, 0] for _ in range(2 - len(res))]

        image_info.roi.colormap = Colormap(colors=res, interpolation=ColormapInterpolationMode.ZERO)
        max_val = image_info.roi_count + 1
        image_info.roi._all_vals = np.array(  # pylint: disable=protected-access
            [0] + [(x + 1) / (max_val + 1) for x in range(1, max_val)]
        )

    def update_roi_coloring(self):
        for image_info in self.image_info.values():
            if image_info.roi is None:
                continue
            self.set_roi_colormap(image_info)
            image_info.roi.opacity = self.settings.get_from_profile(f"{self.name}.image_state.opacity", 1.0)

    def remove_all_roi(self):
        for image_info in self.image_info.values():
            if image_info.roi is None:
                continue
            image_info.roi.visible = False

    def update_roi_border(self) -> None:
        for image_info in self.image_info.values():
            if image_info.roi is None:
                continue
            roi = image_info.roi_info.alternative.get(self.roi_alternative_selection, image_info.roi_info.roi)
            border_thick = self.settings.get_from_profile(f"{self.name}.image_state.border_thick", 1)
            only_border = self.settings.get_from_profile(f"{self.name}.image_state.only_border", True)
            alternative = image_info.roi.metadata.get("alternative", self.roi_alternative_selection)
            if alternative != self.roi_alternative_selection:
                image_info.roi.data = roi
            image_info.roi.contour = border_thick if only_border else 0
            image_info.roi.metadata["alternative"] = self.roi_alternative_selection

    @ensure_main_thread
    def update_rendering(self):
        rendering = self.settings.get_from_profile(RENDERING_MODE_NAME_STR, RENDERING_LIST[0])
        for image_info in self.image_info.values():
            if image_info.roi is not None and hasattr(image_info.roi, "rendering"):
                image_info.roi.rendering = rendering

    @ensure_main_thread
    def update_roi_labeling(self):
        for image_info in self.image_info.values():
            if image_info.roi is not None:
                self.set_roi_colormap(image_info)

    def add_roi_layer(self, image_info: ImageInfo):
        if image_info.roi_info.roi is None:
            return
        roi = image_info.roi_info.alternative.get(self.roi_alternative_selection, image_info.roi_info.roi)
        border_thick = self.settings.get_from_profile(f"{self.name}.image_state.border_thick", 1)
        kwargs = {
            "scale": image_info.image.normalized_scaling(),
            "name": "ROI",
            "blending": "translucent",
            "metadata": {"alternative": self.roi_alternative_selection},
            "rendering": self.settings.get_from_profile(RENDERING_MODE_NAME_STR, RENDERING_LIST[0]),
        }

        only_border = self.settings.get_from_profile(f"{self.name}.image_state.only_border", True)
        image_info.roi = self.viewer.add_labels(roi, **kwargs)
        image_info.roi.contour = border_thick if only_border else 0

    def set_mask(self, mask: Optional[np.ndarray] = None, image: Optional[Image] = None) -> None:
        image = self.get_image(image)
        if image.file_path not in self.image_info:
            raise ValueError("Image not added to viewer")
        if mask is None:
            mask = image.mask

        image_info = self.image_info[image.file_path]
        if mask is None:
            if image_info.mask is not None:
                image_info.mask.visible = False
                image_info.mask.metadata["valid"] = False
                self._toggle_mask_chk_visibility()
            return
        mask_marker = mask == 0
        if image_info.mask is None:
            image_info.mask = self.viewer.add_labels(
                mask_marker, scale=image.normalized_scaling(), blending="translucent", name="Mask"
            )
        else:
            image_info.mask.data = mask_marker
        image_info.mask.metadata["valid"] = True
        if _napari_ge_4_19:
            from napari.utils.colormaps import DirectLabelColormap

            image_info.mask.colormap = DirectLabelColormap(color_dict=self.mask_color())
        else:
            image_info.mask.color = self.mask_color()
        image_info.mask.opacity = self.mask_opacity()
        image_info.mask.visible = self.mask_chk.isChecked()
        self._toggle_mask_chk_visibility()

    def _toggle_mask_chk_visibility(self):
        visibility = any(
            image_info.mask is not None and image_info.mask.metadata.get("valid", False)
            for image_info in self.image_info.values()
        )
        self.mask_chk.setVisible(visibility)
        self.mask_label.setVisible(visibility)

    def update_mask_parameters(self):
        opacity = self.mask_opacity()
        colormap = self.mask_color()
        for image_info in self.image_info.values():
            if image_info.mask is not None:
                image_info.mask.opacity = opacity
                if _napari_ge_4_19:
                    from napari.utils.colormaps import DirectLabelColormap

                    image_info.mask.colormap = DirectLabelColormap(color_dict=colormap)
                else:
                    image_info.mask.color = colormap

    def set_image(self, image: Optional[Image] = None):
        self.image_info = {}
        self.add_image(image, True)

    def has_image(self, image: Image):
        return image.file_path in self.image_info

    @staticmethod
    def calculate_filter(array: np.ndarray, parameters: tuple[NoiseFilterType, float]) -> Optional[np.ndarray]:
        if parameters[0] == NoiseFilterType.No or parameters[1] == 0:
            return array
        if parameters[0] == NoiseFilterType.Gauss:
            return gaussian(array, parameters[1])
        if parameters[0] == NoiseFilterType.Bilateral:
            return bilateral(array, parameters[1])
        return median(array, int(parameters[1]))

    def _remove_worker(self, sender=None):
        if sender is None:
            sender = self.sender()
        for worker in self.worker_list:
            if hasattr(worker, "signals") and sender is worker.signals:
                self.worker_list.remove(worker)
                break
        else:
            logging.debug("[_remove_worker] %s", sender)

    def _add_layer_util(self, index, layer, filters):
        if layer not in self.viewer.layers:
            self.viewer.add_layer(layer)

        if filters[index][0] == NoiseFilterType.No or filters[index][1] == 0:
            return

        worker = calc_layer_filter(layer, filters[index][0], filters[index][1])
        worker.returned.connect(self._add_layer_util_end)
        worker.finished.connect(self._remove_worker)
        self.worker_list.append(worker)
        worker.start()

    @Slot(object)
    def _add_layer_util_end(self, val):
        data_, layer_ = val
        if data_ is not None:
            layer_.data = data_

    def _add_or_move_layer(self, layer: Optional[Layer], index):
        if layer is None:
            return
        if layer not in self.viewer.layers:
            self.viewer.add_layer(layer)
        else:
            self.viewer.layers.move(self.viewer.layers.index(layer), index)

    def _add_image(self, image_data: tuple[ImageInfo, bool]):
        image_info, replace = image_data
        image = image_info.image

        if replace:
            for layer in list(reversed(self.viewer.layers)):
                self.viewer.layers.remove(layer)
            QApplication.instance().processEvents()

        filters = self.channel_control.get_filter()
        for i, layer in enumerate(image_info.layers):
            try:
                self._add_layer_util(i, layer, filters)
            except AssertionError:  # noqa: PERF203
                layer.colormap = "gray"
                self._add_layer_util(i, layer, filters)

        self.image_info[image.file_path].filter_info = filters
        self.image_info[image.file_path].layers = image_info.layers
        self.current_image = image.file_path
        mask_layer = self.image_info[image.file_path].mask
        self._add_or_move_layer(mask_layer, -1)
        roi_layer = self.image_info[image.file_path].roi
        self._add_or_move_layer(roi_layer, -1)
        self.viewer.reset_view()
        if self.viewer.layers:
            if hasattr(self.viewer.layers, "selection"):
                self.viewer.layers.selection.clear()
                self.viewer.layers.selection.add(self.viewer.layers[-1])
            else:
                self.viewer.layers[-1].selected = True

        for i, axis in enumerate(image.array_axis_order):
            if axis == "C":
                continue
            self.viewer.dims.set_point(i, image.shape[i] * image.normalized_scaling()[i] // 2)
        if self.image_info[image.file_path].roi is not None:
            self.set_roi()
        if image_info.image.mask is not None:
            self.set_mask()
        self._toggle_mask_chk_visibility()
        self.image_added.emit()

    def add_image(self, image: Optional[Image], replace=False):
        if image is None:
            image = self.settings.image

        if not image.channels:
            raise ValueError("Need non empty image")

        if image.file_path in self.image_info:
            raise ValueError("Image already added")

        self.image_info[image.file_path] = ImageInfo(image, [])

        channels = image.channels
        if self.image_info and not replace:
            channels = max(channels, *(x.image.channels for x in self.image_info.values()))

        self.channel_control.set_channels(channels)
        visibility = self.channel_control.channel_visibility
        limits = self.channel_control.get_limits()
        ranges = image.get_ranges()
        limits = [ranges[i] if x is None else x for i, x in zip(range(image.channels), limits)]
        gamma = self.channel_control.get_gamma()
        colormaps = [self.channel_control.selected_colormaps[i] for i in range(image.channels)]
        parameters = ImageParameters(
            limits, visibility, gamma, colormaps, image.normalized_scaling(), len(self.viewer.layers)
        )

        self._prepare_layers(image, parameters, replace)

        return image

    if USE_THREADS:  # pragma: no cover

        def _prepare_layers(self, image, parameters, replace):
            worker = prepare_layers(image, parameters, replace)
            worker.returned.connect(self._add_image)
            worker.finished.connect(self._remove_worker)
            self.worker_list.append(worker)
            worker.start()

    else:  # pragma: no cover

        def _prepare_layers(self, image, parameters, replace):
            self._add_image(_prepare_layers(image, parameters, replace))

    def images_bounds(self) -> tuple[list[int], list[int]]:
        ranges = []
        for image_info in self.image_info.values():
            if not image_info.layers:
                continue
            ranges = [
                (min(a, b), max(c, d), min(e, f))
                for (a, c, e), (b, d, f) in itertools.zip_longest(
                    image_info.layers[0].dims.range, ranges, fillvalue=(np.inf, -np.inf, np.inf)
                )
            ]

        visible = [ranges[i] for i in self.viewer.dims.displayed]
        min_shape, max_shape, _ = zip(*visible)
        size = np.subtract(max_shape, min_shape)
        return size, min_shape

    @staticmethod
    def _shift_layer(layer: Layer, translate_2d):
        translate = [0] * layer.ndim
        translate[-2:] = translate_2d
        layer.translate = translate

    def grid_view(self):
        """Present multiple images in grid view"""
        n_row = np.ceil(np.sqrt(len(self.image_info))).astype(int)
        n_row = max(1, n_row)
        scene_size, _ = self.images_bounds()
        for image_info, pos in zip(self.image_info.values(), itertools.product(range(n_row), repeat=2)):
            translate_2d = np.multiply(scene_size[-2:], pos)
            for layer in image_info.layers:
                self._shift_layer(layer, translate_2d)

            if image_info.mask is not None:
                self._shift_layer(image_info.mask, translate_2d)

            if image_info.roi is not None:
                self._shift_layer(image_info.roi, translate_2d)
        self.viewer.reset_view()

    def change_visibility(self, name: str, index: int):
        for image_info in self.image_info.values():
            if len(image_info.layers) > index:
                image_info.layers[index].visible = self.channel_control.channel_visibility[index]
                if self.channel_control.channel_visibility[index]:
                    image_info.layers[index].colormap = self.channel_control.selected_colormaps[index]
                    limits = self.channel_control.get_limits()[index]
                    limits = image_info.image.get_ranges()[index] if limits is None else limits
                    image_info.layers[index].contrast_limits = limits
                    image_info.layers[index].gamma = self.channel_control.get_gamma()[index]
                    filter_type = self.channel_control.get_filter()[index]
                    if filter_type != image_info.filter_info[index]:
                        image_info.layers[index].data = self.calculate_filter(
                            image_info.image.get_channel(index), filter_type
                        )
                        image_info.filter_info[index] = filter_type

    def reset_image_size(self):
        self.viewer.reset_view()

    def set_theme(self, theme: str):
        self.viewer.theme = theme

    def closeEvent(self, event):
        for worker in self.worker_list:
            worker.quit()
        self.viewer.layers.clear()
        self.viewer_widget.close()
        super().closeEvent(event)

    def get_tool_tip_text(self) -> str:
        image = self.settings.image
        image_info = self.image_info[image.file_path]
        text_list = []
        for el in self.components:
            if data := image_info.roi_info.annotations.get(el, {}):
                try:
                    text_list.append(_print_dict(data))
                except ValueError:  # pragma: no cover
                    logging.warning("Wrong value provided as layer annotation.")
        return " ".join(text_list)

    def event(self, event: QEvent):
        if event.type() == QEvent.Type.ToolTip and self.components and (text := self.get_tool_tip_text()):
            QToolTip.showText(event.globalPos(), text, self)
        return super().event(event)

    def _search_component(self):
        try:
            max_components = max(max(image_info.roi_info.bound_info) for image_info in self.image_info.values())
        except ValueError as e:
            if "empty" in e.args[0]:
                return
            raise e
        if self.viewer.dims.ndisplay == 3:
            self._search_type = SearchType.Highlight

        dial = SearchComponentModal(self, self._search_type, self._last_component, max_components)
        if self.viewer.dims.ndisplay == 3:
            dial.zoom_to.setDisabled(True)
        dial.show_right_of_mouse()

    def component_unmark(self, _num):
        for el in self.image_info.values():
            if el.highlight is None:
                continue
            if "timer" in el.highlight.metadata:
                el.highlight.metadata["timer"].stop()
            el.highlight.visible = False

    def _mark_layer(self, num: int, flash: bool, image_info: ImageInfo):
        bound_info = image_info.roi_info.bound_info.get(num, None)
        if bound_info is None:
            return
        # TODO think about marking on bright background
        slices = bound_info.get_slices(1)
        slices[image_info.image.stack_pos] = slice(None)
        component_mark = image_info.roi_info.roi[tuple(slices)] == num
        if self.viewer.dims.ndisplay == 3:
            component_mark = binary_dilation(component_mark)
        shift_base = bound_info.lower - 1
        shift_base[0] += 1  # remove shift on time axis
        translate = image_info.roi.translate + shift_base * image_info.roi.scale
        translate[image_info.image.stack_pos] = 0
        if image_info.highlight is None:
            active_layer = self.viewer.layers.selection.active
            image_info.highlight = self.viewer.add_labels(
                component_mark,
                scale=image_info.roi.scale,
                blending="translucent",
                opacity=0.7,
                **get_highlight_colormap(),
            )
            self.viewer.layers.selection.active = active_layer
        else:
            image_info.highlight.data = component_mark
        image_info.highlight.translate = translate
        image_info.highlight.visible = True
        if flash:
            layer = image_info.highlight

            def flash_fun(layer_=layer):
                opacity = layer_.opacity + 0.1
                if opacity > 1:
                    opacity = 0.1
                layer_.opacity = opacity

            timer = QTimer()
            timer.setInterval(100)
            timer.timeout.connect(flash_fun)
            timer.start()
            layer.metadata["timer"] = timer

    def component_mark(self, num: int, flash: bool = False):
        self.component_unmark(num)
        self._search_type = SearchType.Highlight
        self._last_component = num

        bounding_box = self._bounding_box(num)
        if bounding_box is None:
            return

        for image_info in self.image_info.values():
            self._mark_layer(num, flash, image_info)

        lower_bound, upper_bound = bounding_box
        self._update_point(lower_bound, upper_bound)

        if self.viewer.dims.ndisplay == 2:
            l_bound = lower_bound[-2:][::-1]
            u_bound = upper_bound[-2:][::-1]
            rect = Rect(self.viewer_widget.view.camera.get_state()["rect"])
            if rect.contains(*l_bound) and rect.contains(*u_bound):
                return
            size = u_bound - l_bound
            rect.size = tuple(np.max([rect.size, size * 1.2], axis=0))
            pos = rect.pos
            if rect.left > l_bound[0]:
                pos = l_bound[0], pos[1]
            if rect.right < u_bound[0]:
                pos = pos[0] + u_bound[0] - rect.right, pos[1]
            if rect.bottom > l_bound[1]:
                pos = pos[0], l_bound[1]
            if rect.top < u_bound[1]:
                pos = pos[0], pos[1] + (u_bound[1] - rect.top)
            rect.pos = pos
            self.viewer_widget.view.camera.set_state({"rect": rect})

    def component_zoom(self, num):
        self.component_unmark(num)
        self._search_type = SearchType.Zoom_in
        self._last_component = num

        bounding_box = self._bounding_box(num)
        if bounding_box is None:
            return

        lower_bound, upper_bound = bounding_box
        diff = upper_bound - lower_bound
        frame = diff * (self.settings.get_from_profile(SEARCH_ZOOM_FACTOR_STR, 1.2) - 1)
        if self.viewer.dims.ndisplay == 2:
            rect = Rect(pos=(lower_bound - frame)[-2:][::-1], size=(diff + 2 * frame)[-2:][::-1])
            self.set_state({"camera": {"rect": rect}})
        self._update_point(lower_bound, upper_bound)

    def _update_point(self, lower_bound, upper_bound):
        point = (lower_bound + upper_bound) / 2
        current_point = self.viewer.dims.point
        for i in range(self.viewer.dims.ndim - self.viewer.dims.ndisplay):
            if not (lower_bound[i] <= current_point[i] <= upper_bound[i]):
                self.viewer.dims.set_point(i, point[i])

    @staticmethod
    def _data_to_world(layer: Layer, cords):
        return layer._transforms[1:3].simplified(cords)  # pylint: disable=protected-access

    def _bounding_box(self, num) -> Optional[tuple[np.ndarray, np.ndarray]]:
        lower_bound_list = []
        upper_bound_list = []
        for image_info in self.image_info.values():
            bound_info = image_info.roi_info.bound_info.get(num, None)
            if bound_info is None:
                continue
            lower_bound_list.append(self._data_to_world(image_info.roi, bound_info.lower))
            upper_bound_list.append(self._data_to_world(image_info.roi, bound_info.upper))

        if not lower_bound_list:
            return None

        lower_bound = np.min(lower_bound_list, axis=0)
        upper_bound = np.min(upper_bound_list, axis=0)
        return lower_bound, upper_bound


class NapariQtViewer(QtViewer):
    def __init__(self, viewer):
        super().__init__(viewer, show_welcome_screen=False)
        self.widget(0).layout().setContentsMargins(0, 5, 0, 2)

    def dragEnterEvent(self, event):  # pylint: disable=no-self-use
        """
        ignore napari reading mechanism
        """
        event.ignore()

    def close(self):
        self.dockConsole.deleteLater()
        self.dockLayerList.deleteLater()
        self.dockLayerControls.deleteLater()
        if hasattr(self, "activityDock"):
            self.activityDock.deleteLater()
        return super().close()

    def closeEvent(self, event):
        self.close()
        super().closeEvent(event)

    def _render(self):
        if _napari_ge_5:
            return self.canvas._scene_canvas.render()
        return self.canvas.render()

    if _napari_ge_5:

        @property
        def view(self):
            return self.canvas.view


class SearchComponentModal(QtPopup):
    def __init__(self, image_view: ImageView, search_type: SearchType, component_num: int, max_components):
        super().__init__(image_view)
        self.image_view = image_view
        self.zoom_to = QEnumComboBox(self, SearchType)
        self.zoom_to.setCurrentEnum(search_type)
        self.zoom_to.currentEnumChanged.connect(self._component_num_changed)
        self.component_selector = QSpinBox()
        self.component_selector.valueChanged.connect(self._component_num_changed)
        self.component_selector.setMaximum(max_components)
        self.component_selector.setValue(component_num)

        layout = QHBoxLayout()
        layout.addWidget(QLabel("Component:"))
        layout.addWidget(self.component_selector)
        layout.addWidget(QLabel("Selection:"))
        layout.addWidget(self.zoom_to)
        self.frame.setLayout(layout)

    def _component_num_changed(self):
        if self.zoom_to.currentEnum() == SearchType.Highlight:
            self.image_view.component_mark(self.component_selector.value(), flash=True)
        else:
            self.image_view.component_zoom(self.component_selector.value())

    def closeEvent(self, event):
        super().closeEvent(event)
        self.image_view.component_unmark(0)


@dataclass
class ImageParameters:
    limits: list[tuple[float, float]]
    visibility: list[bool]
    gamma: list[float]
    colormaps: list[Colormap]
    scaling: tuple[Union[float, int]]
    layers: int = 0


def _prepare_layers(image: Image, param: ImageParameters, replace: bool) -> tuple[ImageInfo, bool]:
    image_layers = []
    for i in range(image.channels):
        lim = list(param.limits[i])
        if lim[1] == lim[0]:
            lim[1] += 1
        blending = "additive" if i != 0 else "translucent"
        data = image.get_channel(i)

        layer = _NapariImage(
            data,
            colormap=param.colormaps[i],
            visible=param.visibility[i],
            blending=blending,
            scale=param.scaling,
            contrast_limits=lim,
            gamma=param.gamma[i],
            name=f"channel {i}; {param.layers + i}",
        )
        image_layers.append(layer)
    return ImageInfo(image, image_layers, []), replace


prepare_layers = thread_worker(_prepare_layers)


def _calc_layer_filter(layer: NapariImage, filter_type: NoiseFilterType, radius: float):
    if filter_type == NoiseFilterType.No or radius == 0:
        return None, layer
    return ImageView.calculate_filter(layer.data, parameters=(filter_type, radius)), layer


calc_layer_filter = thread_worker(_calc_layer_filter)


def _print_dict(dkt: MutableMapping, indent="") -> str:
    if not isinstance(dkt, MutableMapping):  # pragma: no cover
        logging.error("%s instead of dict passed to _print_dict", type(dkt))
        return indent + str(dkt)
    res = []
    for k, v in dkt.items():
        if isinstance(v, MutableMapping):
            res.append(f'{indent}{k}:\n{_print_dict(v, f"{indent}  ")}')
        else:
            res.append(f"{indent}{k}: {v}")
    return "\n".join(res)


def _print_list(lst: list, prefix: str) -> str:
    if not lst:
        return ""
    if len(lst) == 1:
        return f"{prefix}: {lst[0]}"
    return f"{prefix}s: {lst}"



from dataclasses import dataclass
from io import BytesIO
from typing import Any, ClassVar, Optional, Protocol, Union, runtime_checkable

import numpy as np

from PartSegCore.mask_create import MaskProperty, calculate_mask
from PartSegCore.roi_info import ROIInfo
from PartSegCore.utils import BaseModel, numpy_repr
from PartSegImage import Image


@dataclass
class AdditionalLayerDescription:
    """
    Dataclass

    :ivar numpy.ndarray data: layer data
    :ivar str layer_type: napari layer type
    :ivar str name: layer name
    """

    data: np.ndarray
    layer_type: str
    name: str = ""

    def __repr__(self):
        return (
            f"AdditionalLayerDescription(data={numpy_repr(self.data)},"
            f" layer_type='{self.layer_type}', name='{self.name}')"
        )


class HistoryElement(BaseModel):
    roi_extraction_parameters: dict[str, Any]
    annotations: Optional[dict[int, Any]]
    mask_property: MaskProperty
    arrays: BytesIO

    class Config:
        arbitrary_types_allowed = True

    @classmethod
    def create(
        cls,
        roi_info: ROIInfo,
        mask: Union[np.ndarray, None],
        roi_extraction_parameters: dict,
        mask_property: MaskProperty,
    ):
        if "name" in roi_extraction_parameters:  # pragma: no cover
            raise ValueError("name")
        arrays = BytesIO()
        arrays_dict = {"roi": roi_info.roi}
        arrays_dict.update(roi_info.alternative.items())
        if mask is not None:
            arrays_dict["mask"] = mask

        np.savez_compressed(arrays, **arrays_dict)
        arrays.seek(0)
        return cls(
            roi_extraction_parameters=roi_extraction_parameters,
            mask_property=mask_property,
            arrays=arrays,
            annotations=roi_info.annotations,
        )

    def get_roi_info_and_mask(self) -> tuple[ROIInfo, Optional[np.ndarray]]:
        self.arrays.seek(0)
        seg = np.load(self.arrays)
        self.arrays.seek(0)
        alternative = {name: array for name, array in seg.items() if name not in {"roi", "mask"}}
        roi_info = ROIInfo(seg["roi"], annotations=self.annotations, alternative=alternative)
        mask = seg.get("mask")
        return roi_info, mask


@runtime_checkable
class ProjectInfoBase(Protocol):
    """
    This is base protocol for Project Information.

    :ivar str ~.file_path: path to current preceded file
    :ivar Image ~.image: project image
    :ivar numpy.ndarray ~.segmentation: numpy array representing current project ROI
    :ivar ROIInfo ~.roi_info: segmentation metadata
    :ivar Optional[numpy.ndarray] ~.mask: mask used in project
    :ivar List[HistoryElement] ~.history: history of calculation
    :ivar str errors: information about problems with current project
    """

    file_path: str
    image: Image
    roi_info: ROIInfo = ROIInfo(None)
    additional_layers: ClassVar[dict[str, AdditionalLayerDescription]] = {}
    mask: Optional[np.ndarray] = None
    history: ClassVar[list[HistoryElement]] = []
    errors: str = ""
    points: Optional[np.ndarray] = None

    def get_raw_copy(self):
        """
        Create copy with only image
        """
        raise NotImplementedError

    def get_raw_mask_copy(self):
        raise NotImplementedError

    def is_raw(self):
        raise NotImplementedError

    def is_masked(self):
        raise NotImplementedError


def calculate_mask_from_project(
    mask_description: MaskProperty, project: ProjectInfoBase, components: Optional[list[int]] = None
) -> np.ndarray:
    """
    Function for calculate mask base on MaskProperty.
    This function calls :py:func:`calculate_mask` with arguments from project.

    :param MaskProperty mask_description: information how calculate mask
    :param ProjectInfoBase project: project with information about segmentation
    :param Optional[List[int]] components: If present inform which components
        should be used when calculation mask, otherwise use all.
    :return: new mask
    :rtype: np.ndarray
    """
    try:
        time_axis = project.image.time_pos
    except AttributeError:
        time_axis = None
    return calculate_mask(
        mask_description, project.roi_info.roi, project.mask, project.image.spacing, components, time_axis
    )


class HistoryProblem(Exception):
    pass



import argparse
import getpass
import locale
import os
import platform
import sys
import zlib
from collections.abc import Sequence
from contextlib import suppress
from importlib.metadata import version as package_version
from typing import Optional

import sentry_sdk
import sentry_sdk.serializer
import sentry_sdk.utils
from packaging.version import parse as parse_version

from PartSeg import __version__, state_store
from PartSeg.common_backend.except_hook import my_excepthook
from PartSegCore.utils import safe_repr

SENTRY_GE_1_29 = parse_version(package_version("sentry_sdk")) >= parse_version("1.29.0")


def proper_suffix(val: str):
    """
    check if val contains only alphanumeric characters

    :raise argparse.ArgumentTypeError: on validation error
    """
    if not val or val.isalnum():
        return val
    raise argparse.ArgumentTypeError(f"suffix '{val}' need to contains only alpha numeric characters")


def proper_path(val: str):
    """
    Check if `val` is proper path in current system

    :raise argparse.ArgumentTypeError: on validation error
    """
    if os.path.exists(val) and os.path.isdir(val):
        return val
    try:
        os.makedirs(val)
    except OSError as e:
        raise argparse.ArgumentTypeError(f" Path {val} is not a valid path in this system") from e

    return val


class CustomParser(argparse.ArgumentParser):
    """
    Argument parser with set of predefined flags:

    #. ``--no_report`` - disable error reporting, still showing dialog with information.
       Set :py:data:`.state_store.report_errors`.
    #. ``--no_dialog`` - disable error reporting and showing dialog. Exceptions will be printed on stderr.
       Set :py:data:`.state_store.show_error_dialog`.
    #. ``--no_update`` - disable check for updates on application startup
        Set :py:data:`.state_store.check_for_updates`.
    #. ``--save_suffix`` - add special suffix to directory contains saved state. designed to allow separate projects.
        Set :py:data:`.state_store.save_suffix`.
    #. ``--save_directory`` - set custom save directory. default can be previewed in `Help > State directory`
         from PartSeg main menu, Set :py:data:`.state_store.save_folder`.
    #. ``--inner_plugins`` - designed to hide part plugin from default start.
         Set :py:data:`.state_store.custom_plugin_load`.
    #. ``--develop`` -- for developer purpose. Allow to reload part of Program without restarting. May be unstable.
         Set :py:data:`.state_store.develop`. Base on this :py:class:`PartSeg.common_gui.advanced_tabs.AdvancedWindow`
         constructor add developer tab.

    """

    def __init__(self, *args, **kwargs):
        if "epilog" not in kwargs and len(args) < 4:
            kwargs["epilog"] = "To control server for sentry reporting use PARTSEG_SENTRY_URL environment variable"
        super().__init__(*args, **kwargs)

        self.add_argument("--no_report", action="store_false", help="disable error reporting")
        self.add_argument("--always_report", action="store_true", help="always report errors without asking user")
        self.add_argument("--no_dialog", action="store_false", help="disable error reporting and showing error dialog")
        self.add_argument("--no_update", action="store_false", help="disable check for updates")
        self.add_argument(
            "--save_suffix",
            "--ssuf",
            type=proper_suffix,
            default=[""],
            help="suffix for configuration_directory",
            nargs=1,
            metavar="suffix",
        )
        self.add_argument(
            "--save_directory",
            "--sdir",
            type=proper_path,
            default=[state_store.save_folder],
            help=f"path to custom configuration folder, if not set then {state_store.save_folder} will be used. "
            "Could be customized using 'PARTSEG_SETTINGS_DIR' environment variable.",
            nargs=1,
            metavar="path",
        )
        self.add_argument("--inner_plugins", action="store_true", help=argparse.SUPPRESS)
        self.add_argument("--develop", action="store_true", help=argparse.SUPPRESS)
        self.add_argument("--version", action="version", version=f"%(prog)s {__version__}")

    def parse_args(self, args: Optional[Sequence[str]] = None, namespace: Optional[argparse.Namespace] = None):
        """
        overload of :py:meth:`argparse.ArgumentParser.parse_args`. Set flags like described in class documentation.
        """
        args = super().parse_args(args, namespace)
        if args.always_report and not args.no_report:
            self.error("Cannot set both --always_report and --no_report")
        state_store.always_report = args.always_report
        state_store.report_errors = args.no_report
        state_store.show_error_dialog = args.no_dialog
        state_store.custom_plugin_load = args.inner_plugins
        state_store.check_for_updates = args.no_update
        state_store.develop = args.develop
        state_store.save_suffix = args.save_suffix[0]
        state_store.save_folder = os.path.abspath(
            args.save_directory[0] + (f"_{state_store.save_suffix}" if state_store.save_suffix else "")
        )

        if args.no_report and args.no_dialog and state_store.sentry_url:
            _setup_sentry()
        sys.excepthook = my_excepthook
        with suppress(locale.Error):
            # some bug in reset locale
            # https://stackoverflow.com/questions/68962248/python-setlocale-with-empty-string-default-locale-gives-unsupported-locale-se
            locale.setlocale(locale.LC_NUMERIC, "")
        return args


def _setup_sentry():  # pragma: no cover
    if not state_store.sentry_url:
        state_store.report_errors = False
        return
    sentry_sdk.utils.MAX_STRING_LENGTH = 10**4
    sentry_sdk.utils.DEFAULT_MAX_VALUE_LENGTH = 10**4
    sentry_sdk.serializer.safe_repr = safe_repr
    sentry_sdk.serializer.MAX_DATABAG_BREADTH = 100
    init_kwargs = {"release": f"PartSeg@{__version__}"}
    if SENTRY_GE_1_29:
        init_kwargs["max_value_length"] = 10**4
    sentry_sdk.init(
        state_store.sentry_url,
        **init_kwargs,
    )
    with sentry_sdk.configure_scope() as scope:
        scope.set_user(
            {
                "name": getpass.getuser(),
                "id": zlib.adler32(f"{getpass.getuser()}#{platform.node()}".encode()),
            }
        )
